/*
 * printl6.c
 * Copyright (C) 2006-2008, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define put_space(file, size) \
    fwrite(fill_space, (size), 1, (file))

typedef struct efield_type {
    const char *class;
    const struct efield_type *efield_data;
    unsigned (*count_field) (void *, unsigned), flags, index;
    void (*print_field) (void *, unsigned);
} efield_type;

typedef struct ffield_type {
    const char *class;
    const struct ffield_type *efield_data;
    unsigned (*count_field) (void *, unsigned), flags, index;
    void (*print_field) (void *, unsigned);
} ffield_type;

typedef struct ifield_type {
    char *class;
    const struct ffield_type *efield_data;
} ifield_type;

typedef struct lfield_type {
    const struct ifield_type *ifield_data;
    struct lfield_type *lfield_data;
} lfield_type;

static int chain_ifields(const struct ffield_type *, char *, unsigned,
			 struct lfield_type **);

static void break_ifields(struct lfield_type *);
static void print_ffields(FILE *, const struct ffield_type *);
static void print_qtracks(FILE *, const struct lfield_type *);

static const char fill_space[] = "                                ";

static int
chain_ifields(const struct ffield_type *ffield_data, char *name, unsigned size,
	      struct lfield_type **lfield)
{
    char *seek = NULL, *this = NULL;
    int status = 0;
    unsigned index;

    index = ffield_data->index;
    while (index) {
	const struct ffield_type *efield_data;

	efield_data = ffield_data->efield_data;
	if (efield_data) {
	    if (!this) {
		this = (char *) malloc(size + 2);
		if (!this) {
		    status = -1;
		    break;
		} else {
		    memcpy(this, name, size);
		    seek = this + size;
		    seek[1] = 0;
		}
	    }

	    *seek = index;

	    {
		struct lfield_type *lfield_data;

		lfield_data = (struct lfield_type *)
		    malloc(sizeof(struct lfield_type)
			   + sizeof(struct ifield_type) + size + 2);
		if (!lfield_data) {
		    status = -1;
		    break;
		} else {
		    char *class;
		    struct ifield_type *ifield_data;

		    ifield_data = (struct ifield_type *) (lfield_data + 1);
		    class = (char *) (ifield_data + 1);

		    memcpy(class, this, size + 2);

		    ifield_data->class = class;
		    ifield_data->efield_data = efield_data;

		    lfield_data->ifield_data = ifield_data;
		    lfield_data->lfield_data = *lfield;

		    *lfield = lfield_data;

		    status = chain_ifields
			(efield_data, this, size + 1, lfield);
		    if (status) {
			break;
		    }
		}
	    }
	}

	ffield_data++;
	index = ffield_data->index;
    }

    if (this) {
	free(this);
    }

    return status;
}


static void
break_ifields(struct lfield_type *lfield_data)
{
    struct lfield_type *ifield_data;

    ifield_data = lfield_data;
    while (ifield_data) {
	lfield_data = lfield_data->lfield_data;
	free(ifield_data);
	ifield_data = lfield_data;
    }
}


static void
print_ffields(FILE *file, const struct ffield_type *ffield_data)
{
    unsigned index;

    index = ffield_data->index;
    while (index) {
	{
	    put_space(file, 2);

	    fputc(index, file);

	    {
		const char *class;

		put_space(file, 29);
		class = ffield_data->class;
		if (class) {
		    fputs(class, file);
		}
	    }

	    fputc(00012, file);
	}

	ffield_data++;
	index = ffield_data->index;
    }
}


static void
print_qtracks(FILE *file, const struct lfield_type *lfield_data)
{
    char c, *class;

    class = lfield_data->ifield_data->class;
    c = *class;
    if (c) {
	unsigned count = 0;

	fputc(00045, file);
	fputc(c, file);
	++class;
	c = *class;
	while (c) {
	    fputc(00050, file);
	    fputc(00045, file);
	    fputc(c, file);
	    ++class;
	    c = *class;
	    count++;
	}
	while (count) {
	    count--;
	    fputc(00051, file);
	}
    }
}


void
x1f4_lprint_ffields6(FILE *file, const struct ffield_type *ffield_data)
{
    int status;
    struct lfield_type *lfield_data = NULL;

    status = chain_ifields(ffield_data, "", 0, &lfield_data);
    if (!status) {
	print_ffields(file, ffield_data);

	while (lfield_data) {
	    struct lfield_type *dfield_data = NULL, *jfield_data, *qfield_data,
		*zfield_data;

	    jfield_data = lfield_data->lfield_data;
	    qfield_data = lfield_data;
	    while (jfield_data) {
		if (strcmp(jfield_data->ifield_data->class,
			   lfield_data->ifield_data->class) < 0) {
		    dfield_data = qfield_data;
		}

		qfield_data = jfield_data;
		jfield_data = jfield_data->lfield_data;
	    }

	    if (dfield_data) {
		jfield_data = dfield_data->lfield_data;
		dfield_data->lfield_data = jfield_data->lfield_data;
	    } else {
		jfield_data = lfield_data;
		lfield_data = jfield_data->lfield_data;
	    }

	    qfield_data = jfield_data;
	    dfield_data = lfield_data;
	    zfield_data = NULL;
	    while (dfield_data) {
		if (dfield_data->ifield_data->efield_data
		    == jfield_data->ifield_data->efield_data) {
		    qfield_data->lfield_data = dfield_data;
		    qfield_data = dfield_data;
		    dfield_data = dfield_data->lfield_data;
		    if (!zfield_data) {
			lfield_data = dfield_data;
		    } else {
			zfield_data->lfield_data = dfield_data;
		    }
		} else {
		    zfield_data = dfield_data;
		    dfield_data = dfield_data->lfield_data;
		}
	    }

	    qfield_data->lfield_data = NULL;

	    dfield_data = jfield_data;
	    qfield_data = jfield_data->lfield_data;
	    while (qfield_data) {
		struct lfield_type *vfield_data;

		vfield_data = qfield_data;
		zfield_data = qfield_data->lfield_data;
		while (zfield_data) {
		    if (strcmp(zfield_data->ifield_data->class,
			       qfield_data->ifield_data->class) < 0) {
			if (vfield_data == qfield_data) {
			    qfield_data->lfield_data =
				zfield_data->lfield_data;
			    zfield_data->lfield_data = qfield_data;
			    dfield_data->lfield_data = zfield_data;
			    qfield_data = zfield_data;
			    zfield_data = vfield_data;
			    vfield_data = qfield_data;
			} else {
			    struct lfield_type *sfield_data;

			    sfield_data = zfield_data->lfield_data;
			    zfield_data->lfield_data =
				qfield_data->lfield_data;
			    qfield_data->lfield_data = sfield_data;
			    dfield_data->lfield_data = zfield_data;
			    vfield_data->lfield_data = qfield_data;
			    sfield_data = zfield_data;
			    zfield_data = qfield_data;
			    qfield_data = sfield_data;
			}
		    }

		    vfield_data = zfield_data;
		    zfield_data = zfield_data->lfield_data;
		}

		dfield_data = qfield_data;
		qfield_data = qfield_data->lfield_data;
	    }

	    fputc(00012, file);

	    qfield_data = jfield_data;
	    print_qtracks(file, qfield_data);
	    qfield_data = qfield_data->lfield_data;
	    while (qfield_data) {
		fputc(00054, file);
		fputc(00040, file);
		print_qtracks(file, qfield_data);
		qfield_data = qfield_data->lfield_data;
	    }

	    fputc(00012, file);

	    print_ffields(file, jfield_data->ifield_data->efield_data);

	    break_ifields(jfield_data);
	}
    }

    break_ifields(lfield_data);
}
