/*
 * id3v2-id3tag.c
 * Copyright (C) 2002-2007, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <autoconfig.h>

#include <getopt.h>
#include <id3tag.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static int dump_5th(struct id3_tag const *);
static int dump_6th(struct id3_tag const *);
static int dump_id3(struct id3_tag const *);
static int dump_tag(struct id3_tag const *, char const *);

static void usage(void);

static int
dump_5th(struct id3_tag const *tag)
{
    struct id3_frame const *frame;

    frame = id3_tag_findframe(tag, ID3_FRAME_COMMENT, 0);
    if (frame) {
	if (1) {
	    id3_ucs4_t const *ucs4;

	    ucs4 = id3_field_getfullstring(&frame->fields[3]);
	    if (ucs4) {
		id3_latin1_t *latin1;

		latin1 = id3_ucs4_latin1duplicate(ucs4);
		if (latin1) {
		    fwrite(latin1, strlen((char *) latin1), 1, stdout);
		    free (latin1);
		}
	    }
	}
    }

    putchar('\0');

    return 0;
}


static int
dump_6th(struct id3_tag const *tag)
{
    struct id3_frame const *frame;

    frame = id3_tag_findframe(tag, ID3_FRAME_GENRE, 0);
    if (frame) {
	union id3_field const *field;
	unsigned j, n;

	field = &frame->fields[1];
	n = id3_field_getnstrings(field);
	for (j = 0; j < n; ++j) {
	    id3_ucs4_t const *ucs4;

	    ucs4 = id3_field_getstrings(field, j);
	    if (!ucs4) {
		break;
	    } else {
		id3_latin1_t *latin1;

		latin1 = id3_ucs4_latin1duplicate(id3_genre_name(ucs4));
		if (!latin1) {
		    break;
		} else {
		    fwrite(latin1, strlen((char *) latin1), 1, stdout);
		    free(latin1);
		}
	    }
	}
    }

    return 0;
}


static int
dump_id3(struct id3_tag const *tag)
{
    int status;

    status = dump_tag(tag, ID3_FRAME_ARTIST);
    if (!status) {
	status = dump_tag(tag, ID3_FRAME_YEAR);
	if (!status) {
	    status = dump_tag(tag, ID3_FRAME_ALBUM);
	    if (!status) {
		status = dump_tag(tag, ID3_FRAME_TITLE);
		if (!status) {
		    status = dump_5th(tag);
		    if (!status) {
			status = dump_6th(tag);
			if (!status) {
			}
		    }
		}
	    }
	}
    }

    return status;
}


static int
dump_tag(struct id3_tag const *tag, char const *what)
{
    struct id3_frame const *frame;

    frame = id3_tag_findframe(tag, what, 0);
    if (frame) {
	union id3_field const *field;
	unsigned j, n;

	field = &frame->fields[1];
	n = id3_field_getnstrings(field);
	for (j = 0; j < n; ++j) {
	    id3_ucs4_t const *ucs4;

	    ucs4 = id3_field_getstrings(field, j);
	    if (!ucs4) {
		break;
	    } else {
		id3_latin1_t *latin1;

		latin1 = id3_ucs4_latin1duplicate(ucs4);
		if (!latin1) {
		    break;
		} else {
		    fwrite(latin1, strlen((char *) latin1), 1, stdout);
		    free(latin1);
		}
	    }
	}
    }

    putchar('\0');

    return 0;
}


static void
usage(void)
{
    puts("Usage: id3v2-id3tag [OPTIONS] PATH\n\
Exit successfully?\n\
\n\
Options:\n\
  -q, --quiet			suppress error messages\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int keep_quiet = 0, status;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "quiet",	    0x00,   NULL,   0x71    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "q", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'q':
		keep_quiet = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (argc - optind != 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	status = 1;
    } else {
	struct id3_file *id3struct;

	id3struct = id3_file_open(argv[optind], ID3_FILE_MODE_READONLY);
	if (!id3struct) {
	    status = 1;
	    if (!keep_quiet) {
		fprintf(stderr, "%s: cannot id3 open `%s'\n", argv[0],
			argv[optind]);
	    }
	} else {
	    struct id3_tag *id3tag;

	    id3tag = id3_file_tag(id3struct);
	    if (!id3tag) {
		status = 1;
		if (!keep_quiet) {
		    fprintf(stderr, "%s: cannot id3 tag `%s'\n", argv[0],
			    argv[optind]);
		}
	    } else {
		status = dump_id3(id3tag);
	    }

	    id3_file_close(id3struct);
	}
    }

    return status;
}
