/*
 * Mylene200.c
 * Copyright (C) 2002-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <signal.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#if !defined HAVE_LIBLEWD
# include <basename.h>
#endif				/* !HAVE_LIBLEWD */
#include <defs.h>
#include <inter.h>
#include <state.h>
#include <types.h>

#if 1
# define FLAGS_0			0
#endif				/* 1 */
#if 1
# define FLAGS_1 \
    FLAGS_0 | Mylene200_e1f4_UNIX | Mylene200_e200_UNIX
#endif				/* 1 */
#if __NONBLOCKING_REFLEX_TRANSFER__
# define FLAGS_2 \
    FLAGS_1 | Mylene200_LOCK_REFLEX
#else
# define FLAGS_2			FLAGS_1
#endif				/* __NONBLOCKING_REFLEX_TRANSFER__ */

#define L_FLAGS				FLAGS_2

static void clean_up(void);
static void flush_up(void);
static void printf_error_0(void *, const char *, va_list);
static void printf_error_1(void *, const char *, va_list);
static void usage(void);

static void *global;

static void
clean_up(void)
{
    struct context_state *context_data;
    struct record_type *record_data;

    context_data = global;
    record_data = context_data->record_data;
    while (record_data) {
	record_data->data(record_data->text);

	record_data = record_data->record_data;
    }
}


static void
flush_up(void)
{
    struct context_state *context_data;
    void *progress;

    context_data = global;

    progress = context_data->progress;

    if (context_data->do_clear) {
	context_data->do_clear = 0;
	if (progress) {
	    fputc('\n', progress);
	}
    }

    if (progress) {
	fflush(progress);
    }
}


static void
printf_error_0(void *context, const char *format, va_list ap)
{
}


static void
printf_error_1(void *context, const char *format, va_list ap)
{
    struct context_state *context_data;
    void *progress;

    context_data = context;

    progress = context_data->progress;

    if (context_data->do_clear) {
	context_data->do_clear = 0;
	if (progress) {
	    fputc('\n', progress);
	}
    }

    if (progress) {
	fflush(progress);
    }

    vfprintf(stderr, format, ap);
}


static void
usage(void)
{
    puts("Usage: Mylene200 [OPTIONS] PATH|PORT Mylene1f4.PATH|Mylene1f4.PORT\n\
Interface Mylene1f4 server.\n\
\n\
Options:\n\
  -I				establish an INET server on PORT\n\
  -U				establish an UNIX server on PATH (default)\n\
  -i, --inet			connect to INET Mylene1f4.PORT server\n\
  -q, --quiet			suppress error messages\n\
  -u, --unix			connect to UNIX Mylene1f4.PATH server\n\
				(default)\n\
      --no-reuse-address	reuse no address when binding INET server\n\
      --reconnect-attempts COUNT\n\
				retry to connect COUNT more many times\n\
      --reconnect-timeout WAIT	wait WAIT microseconds before retrying to\n\
				connect (1,000,000 by default)\n\
      --reuse-address		reuse local address when binding INET server\n\
      --server-mode MODE	set file permissions for local TCP controlli\
ng\n\
				server to integral MODE\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_state context;

    context.Mylene200 = x1f4_basename(argv[0]);
    context.attempts = 0;
    context.do_clear = 0;
    context.flags = L_FLAGS;
    context.mode = NULL;
    context.printf_error = printf_error_1;
    context.progress = NULL;
    context.record_data = NULL;
    context.timeout = 1000000;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *reconnect_attempts_specs = NULL, *reconnect_timeout_specs =
	    NULL;
	int fast = ~0;
	unsigned reuse_record = 0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "inet",	    0x00,   NULL,   0x69    },
		{   "no-reuse-address",
				    0x00,   NULL,   0x16    },
		{   "quiet",	    0x00,   NULL,   0x71    },
		{   "reconnect-attempts",
				    0x01,   NULL,   0x18    },
		{   "reconnect-timeout",
				    0x01,   NULL,   0x19    },
		{   "reuse-address",
				    0x00,   NULL,   0x17    },
		{   "server-mode",  0x01,   NULL,   0x0a    },
		{   "unix",	    0x00,   NULL,   0x75    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "IUiqu", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'I':
		context.flags = (context.flags & ~Mylene200_e200_UNIX)
		    | Mylene200_e200_INET;
		break;
	    case 'U':
		context.flags = (context.flags & ~Mylene200_e200_INET)
		    | Mylene200_e200_UNIX;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'i':
		context.flags = (context.flags & ~Mylene200_e1f4_UNIX)
		    | Mylene200_e1f4_INET;
		break;
	    case 'q':
		context.printf_error = printf_error_0;
		break;
	    case 'u':
		context.flags = (context.flags & ~Mylene200_e1f4_INET)
		    | Mylene200_e1f4_UNIX;
		break;
	    case 012:
		context.mode = optarg;
		break;
	    case 026:
		reuse_record = 0;
		break;
	    case 027:
		reuse_record = Mylene200_REUSE_RECORD;
		break;
	    case 030:
		reconnect_attempts_specs = optarg;
		break;
	    case 031:
		reconnect_timeout_specs = optarg;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	context.flags |= reuse_record;

	do {
	    if (reconnect_attempts_specs) {
		status = Mylene200_parse_cardinal
		    (&context, "reconnect attempts", reconnect_attempts_specs,
		     &context.attempts);
		if (status) {
		    break;
		}
	    }

	    if (reconnect_timeout_specs) {
		status = Mylene200_parse_cardinal
		    (&context, "reconnect timeout", reconnect_timeout_specs,
		     &context.timeout);
		if (status) {
		    break;
		}
	    }
	} while (0);
	if (status) {
	    goto bye;
	}
    }

    if (argc - optind != 2) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    }
    {
	context.e1f4_id = argv[optind + 1];
	context.e200_id = argv[optind + 0];

	global = &context;
	status = Mylene200_all_signals(context.Mylene200, flush_up, clean_up);

	signal(SIGPIPE, SIG_IGN);

	status = Mylene200_init_reflex(&context);
	if (status) {
	} else {
	    status = Mylene200_init_server(&context);
	    if (status) {
	    } else {
		status = Mylene200_line_server(&context);

		if (1) {
		    int excess;

		    excess = Mylene200_fini_server(&context);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    }
		}
	    }

	    if (1) {
		int excess;

		excess = Mylene200_fini_reflex(&context);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }
		}
	    }
	}

	signal(SIGPIPE, SIG_DFL);
    }

  bye:
    ;

    return status;
}
