/*
 * sa1.c
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <errno.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <netdb.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>

#include <access.h>
#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <defs.h>
#include <inter.h>

#include <arpa/inet.h>

#include <netinet/in.h>

#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>

static int bind_hack(void *, int, unsigned);
static int bind_inet(void *, int, unsigned);
static int bind_unix(void *, int);
static int init_inet(void *, unsigned);
static int init_mode(void *);
static int init_path(void *);
static int init_port(void *);
static int init_unix(void *, unsigned);
static int init_user(void *);
static int stat_e3e8(void *, const char *);
static int stat_e7d0(void *, const char *);
static int stat_link(void *);
static int stat_net1(void *, const char *);
static int stat_net2(void *, const char *);
static int stat_net4(void *, const char *);

static int
bind_hack(void *context, int server, unsigned port)
{
    int status = 1;

    status = setsockopt
	(server, SOL_SOCKET, SO_REUSEADDR, &status, sizeof(int));
    if (status) {
	Mylene200_fprintf_error
	    (context, "%s: cannot set \"reuse address\" option for the INET "
	     "%d server\n", Mylene200(context), port);
    }

    return status;
}


static int
bind_inet(void *context, int server, unsigned port)
{
    int status;
    struct sockaddr_in address;

    if (FLAGS(context) & Mylene200_REUSE_RECORD) {
	bind_hack(context, server, port);
    }

    address.sin_family = AF_INET;
    address.sin_port = htons(port);
    address.sin_addr.s_addr = htonl(INADDR_ANY);
    status = bind(server, (struct sockaddr *) &address, sizeof(address));

    return status;
}


static int
bind_unix(void *context, int server)
{
    int status;
    struct sockaddr_un address;

    memset(&address, 0, sizeof(struct sockaddr_un));
    address.sun_family = AF_UNIX;
    strcpy(address.sun_path, E200(context));
    status = bind(server, (struct sockaddr *) &address, sizeof(address));

    return status;
}


static int
init_inet(void *context, unsigned port)
{
    int server, status = 1;

    server = socket(PF_INET, SOCK_STREAM, 0);
    if (server == -1) {
	stat_net2(context, "create");
    } else {
	if (!(server < FD_SETSIZE)) {
	    stat_net1(context, "select");
	} else {
	    status = bind_inet(context, server, port);
	    if (status) {
		stat_net2(context, "bind");
	    } else {
		if (1) {
		    status = listen(server, 1);
		    if (status) {
			stat_net2(context, "listen");
		    } else {
			SERVER(context) = server;

			FD_SET(server, &SET_0(context));
		    }
		}
	    }
	}

	if (status) {
#if __USE_SHUTDOWN__
	    shutdown(server, SHUT_RDWR);
#endif				/* __USE_SHUTDOWN__ */
	    close(server);
	}
    }

    return status;
}


static int
init_mode(void *context)
{
    int status;

    if (MODE(context)) {
	status = init_user(context);
    } else {
	status = init_unix(context, 00600);
    }

    return status;
}


static int
init_path(void *context)
{
    int status = 0;

    do {
	{
	    struct sockaddr_un address;
	    unsigned length;

	    length = strlen(E200(context));
	    if (!(length < sizeof(address.sun_path))) {
		Mylene200_fprintf_error
		    (context, "%s: specified path (`%s') is too long (%u byt"
		     "es long, %u bytes too long)\n", Mylene200(context),
		     E200(context), length,
		     length + 1 - sizeof(address.sun_path));

		status = 1;

		break;
	    }
	}

	status = init_mode(context);
    } while (0);

    return status;
}


static int
init_port(void *context)
{
    int status;
    unsigned port;

    status = x1f4_parse_wxcardinal(&port, E200(context), NULL, 0);
    if (status) {
	Mylene200_fprintf_error
	    (context, "%s: cannot parse port specification: `%s'\n",
	     Mylene200(context), E200(context));
    } else {
	if (65535 < port) {
	    Mylene200_fprintf_error
		(context, "%s: port specification (`%u') is outside valid ra"
		 "nge (0 .. 65535)\n", Mylene200(context), port);
	} else {
	    status = init_inet(context, port);
	}
    }

    return status;
}


static int
init_unix(void *context, unsigned mode)
{
    int server, status = 1;

    server = socket(AF_UNIX, SOCK_STREAM, 0);
    if (server == -1) {
	stat_net2(context, "create");
    } else {
	if (!(server < FD_SETSIZE)) {
	    stat_net1(context, "select");
	} else {
	    Mylene200_set_record(context, Mylene200_lose_server, context);

	    status = bind_unix(context, server);
	    if (status) {
		stat_link(context);
	    } else {
		if (chmod(E200(context), mode)) {
		    status = 1;
		    stat_net4(context, "chmod");
		} else {
		    status = listen(server, 1);
		    if (status) {
			stat_net2(context, "listen");
		    } else {
			SERVER(context) = server;

			FD_SET(server, &SET_0(context));
		    }
		}

		if (status) {
		    unlink(E200(context));
		} else {
		}
	    }

	    if (status) {
		Mylene200_get_record(context, Mylene200_lose_server);
	    }
	}

	if (status) {
#if __USE_SHUTDOWN__
	    shutdown(server, SHUT_RDWR);
#endif				/* __USE_SHUTDOWN__ */
	    close(server);
	}
    }

    return status;
}


static int
init_user(void *context)
{
    int status;
    unsigned mode;

    status = x1f4_parse_wxcardinal(&mode, MODE(context), NULL, 0);
    if (status) {
	Mylene200_fprintf_error
	    (context, "%s: cannot parse mode specification: `%s'\n",
	     Mylene200(context), MODE(context));
    } else {
	if ((mode & 00600) != 00600) {
	    status = 1;
	    Mylene200_fprintf_error
		(context, "%s: I am not even considering running `%s' local "
		 "server without read/write permissions (as mode %u requests"
		 ")\n", Mylene200(context), E200(context), mode);
	} else {
	    status = init_unix(context, mode);
	}
    }

    return status;
}


static int
stat_e3e8(void *context, const char *action)
{
    Mylene200_fprintf_error
	(context, "%s: cannot %s controlling server", Mylene200(context),
	 action);

    return 0;
}


static int
stat_e7d0(void *context, const char *action)
{
    stat_e3e8(context, action);
    Mylene200_fprintf_error(context, " `%s'", E200(context));

    return 0;
}


static int
stat_link(void *context)
{
    stat_e7d0(context, "bind");
    Mylene200_fprintf_error(context, ": %s", strerror(errno));
    if (errno == EADDRINUSE) {
	int excess;
	struct stat info;

	excess = stat(E200(context), &info);
	if (excess) {
	} else {
	    Mylene200_fprintf_error(context, ": %s", strerror(EEXIST));
	}
    }

    Mylene200_fprintf_error(context, "\n");

    return 0;
}


static int
stat_net1(void *context, const char *action)
{
    stat_e3e8(context, action);
    Mylene200_fprintf_error(context, "\n");

    return 0;
}


static int
stat_net2(void *context, const char *action)
{
    stat_e3e8(context, action);
    Mylene200_fprintf_error(context, ": %s\n", strerror(errno));

    return 0;
}


static int
stat_net4(void *context, const char *action)
{
    stat_e7d0(context, action);
    Mylene200_fprintf_error(context, ": %s\n", strerror(errno));

    return 0;
}


int
Mylene200_init_server(void *context)
{
    int status;

    FD_ZERO(&SET_0(context));
    FD_ZERO(&SET_1(context));

    if (FLAGS(context) & Mylene200_e200_UNIX) {
	status = init_path(context);
    } else {
	status = init_port(context);
    }

    return status;
}
