/*
 * sc1.c
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <errno.h>
#if __NONBLOCKING_REFLEX_TRANSFER__
# include <fcntl.h>
#endif				/* __NONBLOCKING_REFLEX_TRANSFER__ */
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <netdb.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>

#include <access.h>
#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <defs.h>
#include <inter.h>

#include <arpa/inet.h>

#include <netinet/in.h>

#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/utsname.h>

static int fail_inet(void *);
static int fail_unix(void *);
static int init_inet(void *);
static int init_link(void *, int, const void *);
static int init_unix(void *);
static int link_inet(void *, int, const void *);
static int link_unix(void *, int, const void *);
#if __NONBLOCKING_REFLEX_TRANSFER__
static int lock_reflex(unsigned *, int);
#endif				/* __NONBLOCKING_REFLEX_TRANSFER__ */
static int miss_link(void *, int, unsigned, void *, int (*) (void *));
static int slip_away(void *, int, unsigned, const char *);

static int
fail_inet(void *context)
{
    return Mylene200_fprintf_error(context, "%s", E1f4(context));
}


static int
fail_unix(void *context)
{
    return Mylene200_fprintf_error(context, "`%s'", E1f4(context));
}


static int
init_inet(void *context)
{
    int status;
    unsigned port;

    status = x1f4_parse_wxcardinal(&port, E1f4(context), NULL, 0);
    if (status) {
	Mylene200_fprintf_error
	    (context, "%s: cannot parse port specification: `%s'\n",
	     Mylene200(context), E1f4(context));
    } else {
	if (65535 < port) {
	    Mylene200_fprintf_error
		(context, "%s: port specification (`%u') is outside valid ra"
		 "nge (0 .. 65535)\n", Mylene200(context), port);
	} else {
	    status = init_link(context, PF_INET, &port);
	}
    }

    return status;
}


static int
init_link(void *context, int domain, const void *argv_1_)
{
    int connection, status;

    connection = socket(domain, SOCK_STREAM, 0);
    if (connection == -1) {
	status = 1;
	Mylene200_fprintf_error
	    (context, "%s: cannot socket: %s\n", Mylene200(context),
	     strerror(errno));
    } else {
	if (domain == AF_UNIX) {
	    status = link_unix(context, connection, argv_1_);
	} else {
	    status = link_inet(context, connection, argv_1_);
	}

#if __NONBLOCKING_REFLEX_TRANSFER__
	if (!status) {
	    lock_reflex(&FLAGS(context), connection);
	}
#endif				/* __NONBLOCKING_REFLEX_TRANSFER__ */

	if (!status) {
#if __REFLEX_SELECT__ == __SINGLE_SELECT__
	    status = slip_away
		(context, connection, 16, "slippingaway\0\0\0\0");
#elif __REFLEX_SELECT__ == __DOUBLE_SELECT__
	    status = slip_away
		(context, connection, 16, "slippingaway\0\0\0\1");
#endif				/* __REFLEX_SELECT__ == __SINGLE_SELECT__ */
	}

	if (status) {
	    close(connection);
	} else {
	    REFLEX(context) = connection;
	}
    }

    return status;
}


static int
init_unix(void *context)
{
    int status;

    if (0) {
    } else {
	if (0) {
	} else {
	    status = init_link(context, AF_UNIX, E1f4(context));
	}
    }

    return status;
}


static int
link_inet(void *context, int connection, const void *argv_1_)
{
    int status = 1;
    struct hostent *hostinfo;
    struct sockaddr_in address;
    struct utsname info;

    address.sin_family = AF_INET;
    address.sin_port = htons(*(unsigned *) argv_1_);
    uname(&info);
    hostinfo = gethostbyname(info.nodename);
    if (hostinfo) {
	address.sin_addr = *(struct in_addr *) hostinfo->h_addr;
	status = miss_link
	    (context, connection, sizeof(address), &address, fail_inet);
    } else {
	Mylene200_fprintf_error
	    (context, "%s: cannot get host name: %s\n", Mylene200(context),
	     hstrerror(h_errno));
    }

    return status;
}


static int
link_unix(void *context, int connection, const void *argv_1_)
{
    int status = 1;
    struct sockaddr_un address;

    memset(&address, 0, sizeof(struct sockaddr_un));
    address.sun_family = AF_UNIX;
    strncpy(address.sun_path, argv_1_, sizeof(address.sun_path) - 1);
    if (1) {
	status = miss_link
	    (context, connection, sizeof(address), &address, fail_unix);
    }

    return status;
}


#if __NONBLOCKING_REFLEX_TRANSFER__
static int
lock_reflex(unsigned *bits, int fd)
{
    int flags;

    flags = fcntl(fd, F_GETFL);
    if (flags == -1) {
    } else {
	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1) {
	} else {
	    int class;

	    class = fcntl(fd, F_GETFL);
	    if (class == -1) {
	    } else {
		if (class != (flags | O_NONBLOCK)) {
		} else {
		    *bits &= ~Mylene200_LOCK_REFLEX;
		}
	    }
	}
    }

    return 0;
}
#endif				/* __NONBLOCKING_REFLEX_TRANSFER__ */


static int
miss_link(void *context, int connection, unsigned size, void *data,
	  int (*fail) (void *))
{
    int status;
    unsigned attempts, timeout;

    attempts = ATTEMPTS(context);
    timeout = TIMEOUT(context);

    do {
	status = connect(connection, data, size);
	if (status) {
	    Mylene200_fprintf_error
		(context, "%s: cannot connect ", Mylene200(context));
	    fail(context);
	    Mylene200_fprintf_error(context, ": %s\n", strerror(errno));
	    if (attempts) {
		attempts--;
		Mylene200_fprintf_error
		    (context, "%s: retrying in %u microseconds\n",
		     Mylene200(context), timeout);
		usleep(timeout);
	    } else {
		break;
	    }
	} else {
	    break;
	}
    } while (1);

    return status;
}


static int
slip_away(void *context, int connection, unsigned size, const char *data)
{
    int status;

    while (size) {
	status = write(connection, data, size);
	if (status == -1) {
	    if (errno == EINTR) {
	    } else {
		Mylene200_lose_reflex(context);
		break;
	    }
	} else {
	    if (!status) {
		status = -1;
		Mylene200_fail_reflex(context);
		break;
	    } else {
		size -= status;
		if (!size) {
		    status = 0;
		    break;
		} else {
		    data += status;
		}
	    }
	}
    }

    return status;
}


int
Mylene200_init_reflex(void *context)
{
    int status;

    if (FLAGS(context) & Mylene200_e1f4_INET) {
	status = init_inet(context);
    } else {
	status = init_unix(context);
    }

    return status;
}
