/*
 * audio_coreaudio.c
 * Copyright (C) 2007-2008, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * based on GPL 2 mpg123 audio_coreaudio.c
 */

#include <sys/time.h>

#include <types.h>

#include <CoreServices/CoreServices.h>
#include <AudioUnit/AudioUnit.h>
#include <AudioToolbox/AudioToolbox.h>
#include <errno.h>
#include <unistd.h>

#include "debug.h"

struct anEnv {
	AudioConverterRef converter;
	AudioUnit outputUnit;
	char play;
	int channels;
	int last_buffer;
	int play_done;

	unsigned char * buffer;
	size_t buffer_size;

	struct xaudio_type *xaudio_data;
};

static struct anEnv *env=NULL;



static OSStatus playProc(AudioConverterRef inAudioConverter,
						 UInt32 *ioNumberDataPackets,
                         AudioBufferList *outOutputData,
                         AudioStreamPacketDescription **outDataPacketDescription,
                         void* inClientData)
{
	long n;
	
	if (env->last_buffer) {
		env->play_done = 1;

		return noErr;
	}

	for(n = 0; n < outOutputData->mNumberBuffers; n++)
	{
		struct xaudio_type *xaudio_data;
		unsigned int wanted = *ioNumberDataPackets * env->channels * 2;
		unsigned char *dest;
		unsigned int read;
		unsigned cache, class, count, state, strip;
		void *audio;

		if (env->buffer_size < wanted) {
			debug1("Allocating %d byte sample conversion buffer", wanted);
			env->buffer = realloc( env->buffer, wanted);
			env->buffer_size = wanted;
		}
		dest = env->buffer;

		xaudio_data = env->xaudio_data;

		audio = xaudio_data->audio;
		class = xaudio_data->class;
		count = xaudio_data->count;
		state = *xaudio_data->state;

		if (state == count) {
			state = 0;
		}
		if (class < state + 1) {
			cache = state - class;
			strip = cache;
		} else {
			cache = count + state - class;
			strip = count - class;
		}
		if (cache < wanted) {
			unsigned trail;

			trail = xaudio_data->trail;

			if (trail == count) {
				trail = 0;
			}

			/*
			 * TODO
			 *
			 *.returning -1 won't make stupid core audio quit
			 */
			if (state != trail) {
				warning("Didn't have any audio data in callb"
					"ack (buffer underflow)");

				return -1;
			}

			wanted = cache;
			env->last_buffer = 1;
		}

		if (wanted < strip) {
			memcpy(dest, (char *) audio + class, wanted);
		} else {
			memcpy(dest, (char *) audio + class, strip);
			memcpy((char *) dest + strip, audio, wanted - strip);
		}

		class += wanted;
		if (class < count) {
		} else {
			class -= count;
		}

		read = wanted;

		xaudio_data->class = class;

		outOutputData->mBuffers[n].mDataByteSize = read;
		outOutputData->mBuffers[n].mData = dest;
	}
	
	return noErr; 
}

static OSStatus convertProc(void *inRefCon, AudioUnitRenderActionFlags *inActionFlags,
                            const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber,
                            UInt32 inNumFrames, AudioBufferList *ioData)
{
	OSStatus err= noErr;
	void *inInputDataProcUserData=NULL;
	AudioStreamPacketDescription* outPacketDescription =NULL;
	struct xaudio_type *xaudio_data;

	err = AudioConverterFillComplexBuffer(env->converter, playProc, inInputDataProcUserData, &inNumFrames, ioData, outPacketDescription);

	xaudio_data = env->xaudio_data;
	if (!xaudio_data->ordered) {
		struct timeval tv;

		xaudio_data->ordered = 1;

		gettimeofday(&tv, NULL);

		xaudio_data->m = tv.tv_sec;
		xaudio_data->n = tv.tv_usec;
	}

	return err;
}


int
_aapple_audio_open(struct xaudio_type *xaudio_data)
{
	UInt32 size;
	ComponentDescription desc;
	Component comp;
	AudioStreamBasicDescription inFormat;
	AudioStreamBasicDescription outFormat;
	AURenderCallbackStruct  renderCallback;
	Boolean outWritable;

	if (!env) {
		env = (struct anEnv*)malloc( sizeof( struct anEnv ) );
		if (!env) {
			error("failed to malloc memory for 'struct anEnv'");

			return -1;
		}
	}

	env->play = 0;
	env->buffer = NULL;
	env->buffer_size = 0;
	env->last_buffer = 0;
	env->play_done = 0;

	env->xaudio_data = xaudio_data;

	desc.componentType = kAudioUnitType_Output; 
	desc.componentSubType = kAudioUnitSubType_DefaultOutput;
	desc.componentManufacturer = kAudioUnitManufacturer_Apple;
	desc.componentFlags = 0;
	desc.componentFlagsMask = 0;
	comp = FindNextComponent(NULL, &desc);
	if (comp == NULL) {
		error("FindNextComponent failed");

		return -1;
	}

	if (OpenAComponent(comp, &(env->outputUnit)))  {
		error("OpenAComponent failed");

		return -1;
	}

	if (AudioUnitInitialize(env->outputUnit)) {
		error("AudioUnitInitialize failed");

		return -1;
	}

	AudioUnitGetPropertyInfo(env->outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &size, &outWritable);
	if (AudioUnitGetProperty(env->outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Output, 0, &outFormat, &size)) {
		error("AudioUnitGetProperty(kAudioUnitProperty_StreamFormat) failed");

		return -1;
	}

	if (AudioUnitSetProperty(env->outputUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &outFormat, size)) {
		error("AudioUnitSetProperty(kAudioUnitProperty_StreamFormat) failed");

		return -1;
	}

	env->channels = xaudio_data->c;
	inFormat.mSampleRate = xaudio_data->e;
	inFormat.mChannelsPerFrame = xaudio_data->c;
	inFormat.mBitsPerChannel = 16;
	inFormat.mBytesPerPacket = 2*inFormat.mChannelsPerFrame;
	inFormat.mFramesPerPacket = 1;
	inFormat.mBytesPerFrame = 2*inFormat.mChannelsPerFrame;
	inFormat.mFormatID = kAudioFormatLinearPCM;
#ifdef _BIG_ENDIAN
	inFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked | kLinearPCMFormatFlagIsBigEndian;
#else
	inFormat.mFormatFlags = kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsPacked;
#endif

	memset(&renderCallback, 0, sizeof(AURenderCallbackStruct));
	renderCallback.inputProc = convertProc;
	renderCallback.inputProcRefCon = 0;
	if (AudioUnitSetProperty(env->outputUnit, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Input, 0, &renderCallback, sizeof(AURenderCallbackStruct))) {
		error("AudioUnitSetProperty(kAudioUnitProperty_SetRenderCallback) failed");

		return -1;
	}

	if (xaudio_data->e > 0 && xaudio_data->c >0 ) {
		if (AudioConverterNew(&inFormat, &outFormat, &(env->converter))) {
			error("AudioConverterNew failed");

			return -1;
		}
		if (xaudio_data->c == 1) {
			SInt32 channelMap[2] = { 0, 0 };
			if (AudioConverterSetProperty(env->converter, kAudioConverterChannelMap, sizeof(channelMap), channelMap)) {
				error("AudioConverterSetProperty(kAudioConverterChannelMap) failed");

				return -1;
			}
		}
	}

	return 0;
}


#if 0
int
_aapple_audio_delay(unsigned *cache)
{
	*cache = 0;

	return 0;
}
#endif


int
_aapple_audio_cache(unsigned *cache)
{
	/*
	 * TODO
	 *
	 *.compute this in a better way
	 */
	*cache = 0;

	return 0;
}


int
_aapple_audio_play_samples(struct xaudio_type *xaudio_data, unsigned char *buf,
			   unsigned trans, unsigned block)
{
	unsigned class, count, index, strip;

	if (32768 < trans) {
		trans = 32768;
	}

	class = xaudio_data->class;
	count = xaudio_data->count;
	index = xaudio_data->index;

	if (index < class + 1) {
		strip = class - index;
	} else {
		strip = count + class - index;
	}

	if (strip < (block ? block : 1)) {
		trans = 0;
	} else {
		if (strip < trans) {
			trans = strip;
		}
		if (block) {
			trans -= trans % block;
		}
		index += trans;
		if (index < count) {
		} else {
			index -= count;
		}

		xaudio_data->index = index;
	}

	if (!env->play) {
		unsigned cache, state;

		state = *xaudio_data->state;

		if (state == count) {
			state = 0;
		}
		if (class < state + 1) {
			cache = state - class;
		} else {
			cache = count + state - class;
		}
		if (32768 < cache) {
			if (AudioOutputUnitStart(env->outputUnit)) {
				error("AudioOutputUnitStart failed");

				return -1;
			}

			env->play = 1;
		} else {
			unsigned trail;

			trail = xaudio_data->trail;

			if (trail == count) {
				trail = 0;
			}

			if (state == trail) {
				if (AudioOutputUnitStart(env->outputUnit)) {
					error("AudioOutputUnitStart failed");

					return -1;
				}

				env->play = 1;
			}
		}
	}

	return trans;
}


int
_aapple_audio_close(struct xaudio_type *xaudio_data, int now)
{
	if (env) {
		if (!now) {
			while (!env->play_done && env->play) {
				usleep(16384);
			}
		}

		AudioConverterDispose(env->converter);
		AudioOutputUnitStop(env->outputUnit);
		AudioUnitUninitialize(env->outputUnit);
		CloseComponent(env->outputUnit);

		if (env->buffer) {
			free( env->buffer );
		}

		free(env);
		env = NULL;

		xaudio_data->started = 0;
	}

	return 0;
}

int
_aapple_audio_clip(struct xaudio_type *xaudio_data, int now, unsigned *delay)
{
	if (env) {
		if (now) {
			/*
			 * TODO
			 *
			 *.compute this in a better way
			 */
			*delay = 0;
		} else {
			*delay = 0;
			while (!env->play_done && env->play) {
				usleep(16384);
			}
		}

		if (AudioOutputUnitStop(env->outputUnit)) {
			error("AudioOutputUnitStop failed");
		}

		env->play = 0;

		_aapple_audio_close(xaudio_data, 1);
	}

	return 0;
}
