/*
 * violet.c
 * Copyright (C) 2005-2008, 2014, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <autoconfig.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#if defined HAVE_LIBx1f4i0
# include <libx1f4i0.h>
#endif				/* HAVE_LIBx1f4i0 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#if !defined HAVE_LIBx1f4i0
# include <cardinal-wx.h>
# include <ckeylist.h>
# include <ckeytree.h>
#endif				/* !HAVE_LIBx1f4i0 */
#include <libx1f4i2.h>

#define __INIT_INTERFACE__		0
#define __OPEN_INTERFACE__		1

#define __HERE_INTERFACE__		__OPEN_INTERFACE__

#define DEVICE				"plughw:0,0"
#undef DEVICE
#define DEVICE				"hw:0,0"

#define BURST_SIZE			(1048576 >> 3)

#define CACHE_SIZE			1048576

#define S010_AUDIO			x1f4_S010_AUDIO
#define S16B_AUDIO			x1f4_S16B_AUDIO
#define S16L_AUDIO			x1f4_S16L_AUDIO
#define S16N_AUDIO			x1f4_S16N_AUDIO
#define U010_AUDIO			x1f4_U010_AUDIO
#define U16B_AUDIO			x1f4_U16B_AUDIO
#define U16L_AUDIO			x1f4_U16L_AUDIO
#define U16N_AUDIO			x1f4_U16N_AUDIO

static int file_audio(const char *, int, const char *);
static int list_key_2(void *, const char *, unsigned);
static int play_audio(const char *, const char *, const char *, int, int,
		      unsigned, void *, int *, unsigned *, unsigned *);

static void print_ckey(const struct x1f4_ckeytree_type *, unsigned);
static void stat_progress(struct timeval *, unsigned, unsigned);
static void usage(void);

static struct x1f4_aerror_type aerror;
static const struct x1f4_ckeytree_type class_0[] = {
/* *INDENT-OFF* */
    {	"BE",		{	S16B_AUDIO	},	NULL		},
    {	"LE",		{	S16L_AUDIO	},	NULL		},
    {	"NE",		{	S16N_AUDIO	},	NULL		},
    {	NULL,		{	0		},	NULL		}
/* *INDENT-ON* */
}, class_1[] = {
/* *INDENT-OFF* */
    {	"16",		{	0		},	class_0		},
    {	"8",		{	S010_AUDIO	},	NULL		},
    {	NULL,		{	0		},	NULL		}
/* *INDENT-ON* */
}, class_2[] = {
/* *INDENT-OFF* */
    {	"BE",		{	U16B_AUDIO	},	NULL		},
    {	"LE",		{	U16L_AUDIO	},	NULL		},
    {	"NE",		{	U16N_AUDIO	},	NULL		},
    {	NULL,		{	0		},	NULL		}
/* *INDENT-ON* */
}, class_3[] = {
/* *INDENT-OFF* */
    {	"16",		{	0		},	class_2		},
    {	"8",		{	U010_AUDIO	},	NULL		},
    {	NULL,		{	0		},	NULL		}
/* *INDENT-ON* */
}, class_4[] = {
/* *INDENT-OFF* */
    {	"S",		{	0		},	class_1		},
    {	"U",		{	0		},	class_3		},
    {	NULL,		{	0		},	NULL		}
/* *INDENT-ON* */
}, class_root[] = {
/* *INDENT-OFF* */
    {	"",		{	0		},	class_4		},
    {	NULL,		{	0		},	NULL		}
/* *INDENT-ON* */
};

static int
file_audio(const char *self, int error, const char *path)
{
    char *name;
    unsigned size;

    size = strlen(path);

    name = (char *) malloc(size + 3);
    if (name) {
	*name = '`';
	name++;
	memcpy(name, path, size);
	name += size;
	*name = '\'';
	name[1] = 0;
	name -= size + 1;

	fprintf(stderr, "%s: ", self);
	x1f4_file_xaudio(stderr, error, name, &aerror);
	fprintf(stderr, "\n");

	free(name);
    }

    return -1;
}


static int
list_key_2(void *text, const char *data, unsigned size)
{
    unsigned *list;

    list = text;
    if (*list) {
	*list = 0;
    } else {
	list[1] += 1;
	fputc('|', stdout);
    }

    list[1] += size;
    fwrite(data, size, 1, stdout);

    return 0;
}


static int
play_audio(const char *self, const char *device, const char *path, int fd,
	   int do_progress, unsigned offset, void *xaudio, int *newout,
	   unsigned *cache_out, unsigned *progress_out)
{
    char cache[CACHE_SIZE];
    int new = 0, status;
    struct timeval now;
    unsigned block, off_0 = 0, off_1 = 0, progress;

    x1f4_size_xaudio(xaudio, &block);

    progress = offset;

    while (1) {
	unsigned source, target;

	source = CACHE_SIZE - off_1;
	if (source < (CACHE_SIZE >> 3)) {
	    target = off_1 - off_0;
	    if (target < (CACHE_SIZE >> 3) + 1) {
		if (target) {
		    memmove(cache, cache + off_0, target);
		}
		off_0 = 0;
		off_1 = target;
	    }
	    source = CACHE_SIZE - off_1;
	}
	if (source) {
	    if (BURST_SIZE < source) {
		source = BURST_SIZE;
	    }
	    status = read(fd, cache + off_1, source);
	    if (status == -1) {
		if (new) {
		    new = 0;
		    fprintf(stderr, "\n");
		}
		fprintf(stderr, "%s: cannot read `%s': %s\n", self, path,
			strerror(errno));
		break;
	    } else {
		if (!status) {
		    break;
		} else {
		    off_1 += status;
		    progress += status;
		}
	    }
	}

	target = off_1 - off_0;
	if (block < target + 1) {
	    status = x1f4_play_xaudio(xaudio, cache + off_0, target, &now);
	    if (status < 0) {
		if (new) {
		    new = 0;
		    fprintf(stderr, "\n");
		}
		file_audio(self, status, device);

		break;
	    } else {
		if (!status) {
		    if (new) {
			fprintf(stderr, "\n");
		    }
		    fprintf(stderr, "%s[%u]\n", __FILE__, __LINE__);
		    exit(1);
		} else {
		    off_0 += status;
		}
	    }
	}

	if (do_progress) {
	    stat_progress(&now, off_1 - off_0, progress);
	    new = 1;
	}
    }

    if (!status) {
	while (1) {
	    unsigned target;

	    target = off_1 - off_0;
	    if (target < block) {
		if (!target) {
		    status = 0;
		    break;
		}
		status = x1f4_xlay_xaudio(xaudio, cache + off_0, target, &now);
	    } else {
		status = x1f4_play_xaudio(xaudio, cache + off_0, target, &now);
	    }
	    {
		if (status < 0) {
		    if (new) {
			new = 0;
			fprintf(stderr, "\n");
		    }
		    file_audio(self, status, device);

		    break;
		} else {
		    if (!status) {
			if (new) {
			    fprintf(stderr, "\n");
			}
			fprintf(stderr, "%s[%u]\n", __FILE__, __LINE__);
			exit(1);
		    } else {
			off_0 += status;
		    }
		}
	    }

	    if (do_progress) {
		stat_progress(&now, off_1 - off_0, progress);
		new = 1;
	    }
	}
    }

    if (status) {
	*newout = 0;
    } else {
	*newout = new;
	if (new) {
	    *cache_out = off_1 - off_0;
	    *progress_out = progress;
	    stat_progress(&now, off_1 - off_0, progress);
	}
    }

    return status;
}


static void
print_ckey(const struct x1f4_ckeytree_type *root, unsigned header)
{
    unsigned list[2];

    list[0] = 1;
    list[1] = header;

    x1f4_list_ctkey(list, list_key_2, root);

    if (31 < list[1]) {
	fputs("\n\t\t\t\t", stdout);
    } else {
	unsigned i;

	i = (32 - list[1] + 7) >> 3;
	while (i) {
	    fputc('\t', stdout);
	    i--;
	}
    }
}


static void
stat_progress(struct timeval *now, unsigned cache, unsigned offset)
{
    fprintf(stderr, "\r %15u %15u     %02u:%02u:%02u:%02u", cache, offset,
	    (unsigned) (now->tv_sec / 3600) % 60,
	    (unsigned) (now->tv_sec / 60) % 60, (unsigned) now->tv_sec % 60,
	    (unsigned) (now->tv_usec * 60 / 1000000));
}


static void
usage(void)
{
    fputs("Usage: violet [OPTIONS] [FILE]\n\
Play stereo PCM FILE(s).\n\
\n\
Options:\n\
  -c, --class ", stdout);
    print_ckey(class_root, 14);
    puts("set audio data class (defaults to S16NE)\n\
  -f, --sampling-frequency FREQUENCY\n\
				set sampling frequency (defaults to 44100)\n\
  -o, --offset OFFSET		start playing at this offset\n\
  -p, --progress		display progress information\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    const char *device = NULL;
    int do_progress = 0, status = 0;
    unsigned class = x1f4_S16N_AUDIO, channels = 2, offset = 0,
	sampling_frequency = 44100;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *class_specs = NULL, *offset_specs = NULL,
	    *sampling_frequency_specs = NULL;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "class",	    0x01,   NULL,   0x63    },
		{   "offset",       0x01,   NULL,   0x6f    },
		{   "progress",	    0x00,   NULL,   0x70    },
		{   "sampling-frequency",
				    0x01,   NULL,   0x66    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "c:f:o:p", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'c':
		class_specs = optarg;
		break;
	    case 'f':
		sampling_frequency_specs = optarg;
		break;
	    case 'h':
		usage();

		return 0;
	    case 'o':
		offset_specs = optarg;
		break;
	    case 'p':
		do_progress = 1;
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);

		return 0;
	    case '?':
		return 1;
	    }
	}

	if (class_specs) {
	    if (x1f4_find_ctkey(class_specs, class_root, &class)) {
		fprintf(stderr, "%s: cannot parse audio data class specifica"
			"tion: `%s'\n", argv[0], class_specs);

		return 1;
	    }
	}

	if (offset_specs) {
	    if (x1f4_parse_wxcardinal(&offset, offset_specs, NULL, 0)) {
		fprintf(stderr, "%s: cannot parse offset specification: `%s'"
			"\n", argv[0], offset_specs);

		return 1;
	    }
	}

	if (sampling_frequency_specs) {
	    if (x1f4_parse_wxcardinal
		(&sampling_frequency, sampling_frequency_specs, NULL, 0)) {
		fprintf(stderr, "%s: cannot parse sampling frequency specifi"
			"cation: `%s'\n", argv[0], sampling_frequency_specs);

		return 1;
	    }
	}
    }

    if (!device) {
	device = DEVICE;
    }

#if __HERE_INTERFACE__ == __OPEN_INTERFACE__
    {
	void *xaudio;

	status = x1f4_open_xaudio(&xaudio, device, &aerror);
	if (status) {
	    file_audio(argv[0], status, device);
	} else {
#endif				/* __HERE_INTERFACE__ == __OPEN_INTERFACE__ */
	    for (; optind < argc; optind++) {
		int fd;

		fd = open(argv[optind], O_RDONLY);
		if (fd == -1) {
		    status = 1;
		    fprintf(stderr, "%s: cannot open `%s': %s\n", argv[0],
			    argv[optind], strerror(errno));
		    break;
		} else {
		    if (offset) {
			status = lseek(fd, offset, SEEK_SET);
			if (status != -1) {
			    status = 0;
			}
		    }
		    if (!status) {
#if __HERE_INTERFACE__ == __INIT_INTERFACE__
			void *xaudio;
#endif				/* __HERE_INTERFACE__ == __INIT_INTERFACE__ */

#if __HERE_INTERFACE__ == __INIT_INTERFACE__
			status = x1f4_init_xaudio
			    (&xaudio, device, &aerror, sampling_frequency,
			     channels, class, 0);
#else
			status = x1f4_mode_xaudio
			    (xaudio, sampling_frequency, channels, class, 0);
#endif				/* __HERE_INTERFACE__ == __INIT_INTERFACE__ */
			if (status) {
			    file_audio(argv[0], status, device);
			} else {
			    int new;
			    unsigned cache = 0, progress = 0;

			    if (offset) {
				x1f4_bias_xaudio(xaudio, offset);
			    }

			    status = play_audio
				(argv[0], device, argv[optind], fd,
				 do_progress, offset, xaudio, &new, &cache,
				 &progress);
			    if (status) {
				int excess;

#if __HERE_INTERFACE__ == __INIT_INTERFACE__
				excess = x1f4_fini_xaudio(&xaudio, 1);
				if (excess) {
				    file_audio(argv[0], excess, device);
				}
#else
				excess = x1f4_deck_xaudio(xaudio, 1);
				if (excess) {
				    file_audio(argv[0], excess, device);
				}
#endif				/* __HERE_INTERFACE__ == __INIT_INTERFACE__ */
			    } else {
				struct timeval now;

#if __HERE_INTERFACE__ == __INIT_INTERFACE__
				status = x1f4_stop_xaudio(&xaudio, 0, &now);
#else
				status = x1f4_clip_xaudio(xaudio, 0, &now);
#endif				/* __HERE_INTERFACE__ == __INIT_INTERFACE__ */
				if (status) {
				    if (new) {
					fprintf(stderr, "\n");
				    }
				    file_audio(argv[0], status, device);
				} else {
				    if (new) {
					if (do_progress) {
					    stat_progress
						(&now, cache, progress);
					    fprintf(stderr, "\n");
					}
				    }
				}
			    }
			}
		    }

		    {
			int excess;

			excess = close(fd);
			if (excess) {
			    status = 1;
			    fprintf(stderr, "%s: cannot close `%s': %s\n",
				    argv[0], argv[optind], strerror(errno));
			}
		    }

		    if (status) {
			break;
		    }
		}
	    }

#if __HERE_INTERFACE__ == __OPEN_INTERFACE__
	    {
		int excess;

		excess = x1f4_lose_xaudio(&xaudio);
		if (excess) {
		    status = 1;
		    file_audio(argv[0], excess, device);
		}
	    }
	}
    }
#endif				/* __HERE_INTERFACE__ == __OPEN_INTERFACE__ */

    return status;
}
