/**********************************************************/
/* inverse two dimensional DCT, Chen-Wang algorithm       */
/* (cf. IEEE ASSP-32, pp. 803-816, Aug. 1984)             */
/* 32-bit integer arithmetic (8 bit coefficients)         */
/* 11 mults, 29 adds per DCT                              */
/*                                      sE, 18.8.91       */
/**********************************************************/
/* coefficients extended to 12 bit for IEEE1180-1990      */
/* compliance                           sE,  2.1.94       */
/**********************************************************/
/* short turned to seek_h262_coefficient_type in function */
/* arguments list                                         */
/*                                                        */
/* I found this code in                                   */
/* LIBMPEG3                                               */
/* Author: Adam Williams <broadcast@earthling.net>        */
/* Page: heroine.linuxbox.com                             */
/*                                                        */
/* ciprian niculescu                         6.0.02       */
/**********************************************************/

#include <config.h>

#include <idct.1.h>

#define ARITHMETIC_CLIPPING		0
#define LOOKUP_CLIPPING			1

#if 1
# define __INTRA_CLIPPING__		LOOKUP_CLIPPING
#else
# define __INTRA_CLIPPING__		ARITHMETIC_CLIPPING
#endif				/* 0 */
#if 0
# define __NON_INTRA_CLIPPING__		LOOKUP_CLIPPING
#else
# define __NON_INTRA_CLIPPING__		ARITHMETIC_CLIPPING
#endif				/* 0 */

#define ARITHMETIC_BIAS			256
#define LOOKUP_BIAS			320

#define W1				2841
#define W2				2676
#define W3				2408
#define W5				1609
#define W6				1108
#define W7				565

#define CLIP_LOOKUP			0

#if __INTRA_CLIPPING__ == LOOKUP_CLIPPING
# undef CLIP_LOOKUP
# define CLIP_LOOKUP			1
# define INTRA_IDCT_BIAS		LOOKUP_BIAS
#else
# define INTRA_IDCT_BIAS		ARITHMETIC_BIAS
#endif				/* __INTRA_CLIPPING__ == LOOKUP_CLIPPING */
#if __NON_INTRA_CLIPPING__ == LOOKUP_CLIPPING
# undef CLIP_LOOKUP
# define CLIP_LOOKUP			1
# define NON_INTRA_IDCT_BIAS		LOOKUP_BIAS
#else
# define NON_INTRA_IDCT_BIAS		ARITHMETIC_BIAS
#endif				/* __NON_INTRA_CLIPPING__ == LOOKUP_CLIPPING */

#define ADD(k, c)			(k) += (c)
#define SET(k, c)			(k) = (c)

#if !__SHORT_ENDED_ARITHMETIC__
# define DETACH_ROW(lock) \
    {									      \
	*(lock)++ = 0;							      \
	*(lock)++ = 0;							      \
	*(lock)++ = 0;							      \
	*(lock)++ = 0;							      \
	*(lock)++ = 0;							      \
	*(lock)++ = 0;							      \
	*(lock)++ = 0;							      \
	*(lock)++ = 0;							      \
    }
#elif 0
# define DETACH_ROW(lock) \
    {									      \
	*((unsigned *) lock)++ = 0;					      \
	*((unsigned *) lock)++ = 0;					      \
	*((unsigned *) lock)++ = 0;					      \
	*((unsigned *) lock)++ = 0;					      \
    }
#else
# define DETACH_ROW(lock) \
    {									      \
	unsigned *data;							      \
									      \
	data = (unsigned *) (lock);					      \
	*data++ = 0;							      \
	*data++ = 0;							      \
	*data++ = 0;							      \
	*data++ = 0;							      \
	(lock) = (seek_h262_coefficient_type *) data;			      \
    }
#endif				/* !__SHORT_ENDED_ARITHMETIC__ */

#if CLIP_LOOKUP
typedef unsigned limit_type;
#endif				/* CLIP_LOOKUP */

#if CLIP_LOOKUP
static int range_limit_init = 1;
static limit_type range_limit[2048];
#endif				/* CLIP_LOOKUP */

void
mpeg3video_idct_conversion_0(struct seek_h262_block_type *blocks,
			     unsigned pattern_code, unsigned offset,
			     unsigned stride)
{
    for (; pattern_code; pattern_code >>= 2) {
	if (pattern_code & 1) {
	    {
		seek_h262_coefficient_type *lock;
		unsigned i = 8;
#if __ENABLE_BLOCK_META_DATA__
		unsigned meta_block;
#endif				/* __ENABLE_BLOCK_META_DATA__ */

		lock = blocks->coefficients;

#if __ENABLE_BLOCK_META_DATA__
# if __SHORT_ENDED_ARITHMETIC__
		meta_block = ((unsigned short *) lock)[-1];
# else
		meta_block = ((unsigned *) lock)[-1];
# endif				/* __SHORT_ENDED_ARITHMETIC__ */
		meta_block = ~meta_block << 1;
#endif				/* __ENABLE_BLOCK_META_DATA__ */

		for (; i; i--) {
		    int x0, x1, x2, x3, x4, x5, x6, x7, x8;

#if __ENABLE_BLOCK_META_DATA__
		    meta_block >>= 1;
		    if (meta_block & 1) {
			lock++;
			continue;
		    }
		    if (meta_block & 256) {
			lock[000] = lock[010] = lock[020] = lock[030] =
			    lock[040] = lock[050] = lock[060] = lock[070] =
			    lock[000] << 3;
			lock++;
			continue;
		    }

		    x0 = (lock[000] << 11) + 128;
		    x1 = lock[040] << 11;
		    x2 = lock[060];
		    x3 = lock[020];
		    x4 = lock[010];
		    x5 = lock[070];
		    x6 = lock[050];
		    x7 = lock[030];
#else
		    if (!((x1 = lock[040] << 11) | (x2 = lock[060])
			  | (x3 = lock[020]) | (x4 = lock[010])
			  | (x5 = lock[070]) | (x6 = lock[050])
			  | (x7 = lock[030]))) {
			lock[000] = lock[010] = lock[020] = lock[030] =
			    lock[040] = lock[050] = lock[060] = lock[070] =
			    lock[000] << 3;
			lock++;
			continue;
		    }

		    x0 = (lock[000] << 11) + 128;
#endif				/* __ENABLE_BLOCK_META_DATA__ */

		    x8 = W7 * (x4 + x5);
		    x4 = x8 + (W1 - W7) * x4;
		    x5 = x8 - (W1 + W7) * x5;
		    x8 = W3 * (x6 + x7);
		    x6 = x8 - (W3 - W5) * x6;
		    x7 = x8 - (W3 + W5) * x7;

		    x8 = x0 + x1;
		    x0 -= x1;
		    x1 = W6 * (x3 + x2);
		    x2 = x1 - (W2 + W6) * x2;
		    x3 = x1 + (W2 - W6) * x3;
		    x1 = x4 + x6;
		    x4 -= x6;
		    x6 = x5 + x7;
		    x5 -= x7;

		    x7 = x8 + x3;
		    x8 -= x3;
		    x3 = x0 + x2;
		    x0 -= x2;
		    x2 = (181 * (x4 + x5) + 128) >> 8;
		    x4 = (181 * (x4 - x5) + 128) >> 8;

		    lock[000] = (x7 + x1) >> 8;
		    lock[010] = (x3 + x2) >> 8;
		    lock[020] = (x0 + x4) >> 8;
		    lock[030] = (x8 + x6) >> 8;
		    lock[040] = (x8 - x6) >> 8;
		    lock[050] = (x0 - x4) >> 8;
		    lock[060] = (x3 - x2) >> 8;
		    lock[070] = (x7 - x1) >> 8;

		    lock++;
		}
	    }

	    {
		seek_h262_coefficient_type *lock;
		unsigned i = 8;
		unsigned char *output;

		output = blocks->base_address + offset;

		lock = blocks->coefficients;
		for (; i; i--) {
		    int x0, x1, x2, x3, x4, x5, x6, x7, x8;

#if __SHORT_ENDED_ARITHMETIC__
		    if (!lock[1]
			&& !(((int *) lock)[1] | ((int *) lock)[2]
			     | ((int *) lock)[3])) {
#elif 0
		    }
#else
		    if (!((x1 = lock[4] << 8) | (x2 = lock[6]) | (x3 = lock[2])
			  | (x4 = lock[1]) | (x5 = lock[7]) | (x6 = lock[5])
			  | (x7 = lock[3]))) {
#endif				/* __SHORT_ENDED_ARITHMETIC__ */
#if 0
			limit_type *limit;
			unsigned value;

			value =
			    (lock[0] + 32 + (NON_INTRA_IDCT_BIAS << 6)) >> 6;
			if ((int) value < 0) {
			    limit = range_limit;
			} else if (value < 2048 - 256) {
			    limit = range_limit + value;
			} else {
			    limit = range_limit + 2048 - 256;
			}

			*output++ = limit[*output];
			*output++ = limit[*output];
			*output++ = limit[*output];
			*output++ = limit[*output];
			*output++ = limit[*output];
			*output++ = limit[*output];
			*output++ = limit[*output];
			*output++ = limit[*output];

			output += stride - 8;
# else
			int value;
			unsigned *output_x20;

/*	    if (lock[0] < 0) {
		value = 256 - ((-(lock[0] | ~16383) + 32) >> 6);
		if (value < 0) {
		    value = 0;
		}
	    } else {
		value = 256 + (((lock[0] & 16383) + 32) >> 6);
		if (511 < value) {
		    value = 511;
		}
	    }*/
	    value = (lock[0] + 32) >> 6;
	    if (value < 0) {
		value = 256 + value;
		if (value < 0) {
		    value = 0;
		}
	    } else {
		value = 256 + value;
		if (511 < value) {
		    value = 511;
		}
	    }
			if (value < 256) {
			    unsigned value_x7f7f7f7f;

			    value |= value << 010;
			    value |= value << 020;
			    value_x7f7f7f7f = (unsigned) value >> 1;
			    if (value & 1) {
				unsigned sum_0, sum_1, sum_2;

				value_x7f7f7f7f &= 0x7f7f7f7f;

				output_x20 = (unsigned *) output;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_0 += sum_1;
				sum_1 ^= 0x01010101;
				sum_2 = sum_0 & 0x80808080;
				sum_1 &= sum_2 >> 7;
				sum_1 += (sum_0 & (sum_2 - (sum_2 >> 7))) << 1;
				*output_x20++ = sum_1;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_0 += sum_1;
				sum_1 ^= 0x01010101;
				sum_2 = sum_0 & 0x80808080;
				sum_1 &= sum_2 >> 7;
				sum_1 += (sum_0 & (sum_2 - (sum_2 >> 7))) << 1;
				*output_x20 = sum_1;
			    } else {
				unsigned sum_0, sum_1, sum_2;

				output_x20 = (unsigned *) output;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_2 = sum_0 & 0x80808080;
				sum_1 &= sum_2 >> 7;
				sum_1 += (sum_0 & (sum_2 - (sum_2 >> 7))) << 1;
				*output_x20++ = sum_1;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_2 = sum_0 & 0x80808080;
				sum_1 &= sum_2 >> 7;
				sum_1 += (sum_0 & (sum_2 - (sum_2 >> 7))) << 1;
				*output_x20 = sum_1;
			    }
			} else {
			    unsigned value_x7f7f7f7f;

			    value &= 255;
			    value |= value << 010;
			    value |= value << 020;
			    value_x7f7f7f7f = (unsigned) value >> 1;
			    if (value & 1) {
				unsigned sum_0, sum_1;

				value_x7f7f7f7f &= 0x7f7f7f7f;

				output_x20 = (unsigned *) output;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_0 += sum_1;
				sum_1 ^= 0x01010101;
				if (sum_0 & 0x80808080) {
				    unsigned mask;

				    mask = sum_0 & 0x80808080;
				    mask |= mask - (mask >> 7);
				    *output_x20++ =
					(((sum_0 & 0x7f7f7f7f) << 1) + sum_1)
					| mask;
				} else {
				    *output_x20++ = (sum_0 << 1) + sum_1;
				}
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_0 += sum_1;
				sum_1 ^= 0x01010101;
				if (sum_0 & 0x80808080) {
				    unsigned mask;

				    mask = sum_0 & 0x80808080;
				    mask |= mask - (mask >> 7);
				    *output_x20 =
					(((sum_0 & 0x7f7f7f7f) << 1) + sum_1)
					| mask;
				} else {
				    *output_x20 = (sum_0 << 1) + sum_1;
				}
			    } else {
				unsigned sum_0, sum_1;

				output_x20 = (unsigned *) output;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				if (sum_0 & 0x80808080) {
				    unsigned mask;

				    mask = sum_0 & 0x80808080;
				    mask |= mask - (mask >> 7);
				    *output_x20++ =
					(((sum_0 & 0x7f7f7f7f) << 1) + sum_1)
					| mask;
				} else {
				    *output_x20++ = (sum_0 << 1) + sum_1;
				}
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				if (sum_0 & 0x80808080) {
				    unsigned mask;

				    mask = sum_0 & 0x80808080;
				    mask |= mask - (mask >> 7);
				    *output_x20 =
					(((sum_0 & 0x7f7f7f7f) << 1) + sum_1)
					| mask;
				} else {
				    *output_x20 = (sum_0 << 1) + sum_1;
				}
			    }
			}

			output += stride;
#endif				/* 0 */

			DETACH_ROW(lock);

			continue;
		    }

#if __SHORT_ENDED_ARITHMETIC__
		    x1 = lock[4] << 8;
		    x2 = lock[6];
		    x3 = lock[2];
		    x4 = lock[1];
		    x5 = lock[7];
		    x6 = lock[5];
		    x7 = lock[3];
#endif				/* __SHORT_ENDED_ARITHMETIC__ */

		    x0 = (lock[0] << 8) + 8192 + (NON_INTRA_IDCT_BIAS << 14);

		    DETACH_ROW(lock);

		    x8 = W7 * (x4 + x5) + 4;
		    x4 = (x8 + (W1 - W7) * x4) >> 3;
		    x5 = (x8 - (W1 + W7) * x5) >> 3;
		    x8 = W3 * (x6 + x7) + 4;
		    x6 = (x8 - (W3 - W5) * x6) >> 3;
		    x7 = (x8 - (W3 + W5) * x7) >> 3;

		    x8 = x0 + x1;
		    x0 -= x1;
		    x1 = W6 * (x3 + x2) + 4;
		    x2 = (x1 - (W2 + W6) * x2) >> 3;
		    x3 = (x1 + (W2 - W6) * x3) >> 3;
		    x1 = x4 + x6;
		    x4 -= x6;
		    x6 = x5 + x7;
		    x5 -= x7;

		    x7 = x8 + x3;
		    x8 -= x3;
		    x3 = x0 + x2;
		    x0 -= x2;
		    x2 = (181 * (x4 + x5) + 128) >> 8;
		    x4 = (181 * (x4 - x5) + 128) >> 8;

#if __NON_INTRA_CLIPPING__ == LOOKUP_CLIPPING
		    *output++ =
			range_limit[(((x7 + x1) >> 14) + *output) & 2047];
		    *output++ =
			range_limit[(((x3 + x2) >> 14) + *output) & 2047];
		    *output++ =
			range_limit[(((x0 + x4) >> 14) + *output) & 2047];
		    *output++ =
			range_limit[(((x8 + x6) >> 14) + *output) & 2047];
		    *output++ =
			range_limit[(((x8 - x6) >> 14) + *output) & 2047];
		    *output++ =
			range_limit[(((x0 - x4) >> 14) + *output) & 2047];
		    *output++ =
			range_limit[(((x3 - x2) >> 14) + *output) & 2047];
		    *output++ =
			range_limit[(((x7 - x1) >> 14) + *output) & 2047];

		    output += stride - 8;
#else
		    {
			unsigned c, k;

			SET(k, (((x7 + x1) >> 14) + output[0]) << 000);
			ADD(k, (((x3 + x2) >> 14) + output[1]) << 012);
			ADD(k, (((x0 + x4) >> 14) + output[2]) << 024);
			c = k & 0x20080200;
			k = k | (c - (c >> 9));
			c = k & 0x10040100;
			k = k & (c - (c >> 8));

			output[0] = k >> 000;
			output[1] = k >> 012;
			output[2] = k >> 024;

			SET(k, (((x8 + x6) >> 14) + output[3]) << 000);
			ADD(k, (((x8 - x6) >> 14) + output[4]) << 012);
			c = k & 0x00080200;
			k = k | (c - (c >> 9));
			c = k & 0x00040100;
			k = k & (c - (c >> 8));

			output[3] = k >> 000;
			output[4] = k >> 012;

			SET(k, (((x0 - x4) >> 14) + output[5]) << 000);
			ADD(k, (((x3 - x2) >> 14) + output[6]) << 012);
			ADD(k, (((x7 - x1) >> 14) + output[7]) << 024);
			c = k & 0x20080200;
			k = k | (c - (c >> 9));
			c = k & 0x10040100;
			k = k & (c - (c >> 8));

			output[5] = k >> 000;
			output[6] = k >> 012;
			output[7] = k >> 024;
		    }

		    output += stride;
#endif				/* __NON_INTRA_CLIPPING__ == LOOKUP_CLIPPING */
		}
	    }
	} else if (pattern_code & 2) {
	    {
#if 0
		limit_type *limit;
		seek_h262_coefficient_type block;
		unsigned i = 8;
		unsigned char *output;

		output = blocks->base_address + offset;

		limit = (limit_type *) range_limit;
		block = *blocks->coefficients;
		*blocks->coefficients = 0;
		if (block < 0) {
		    limit +=
			NON_INTRA_IDCT_BIAS - ((-(block | ~2047) + 4) >> 3);
		} else {
		    limit += NON_INTRA_IDCT_BIAS + (((block & 2047) + 4) >> 3);
		}

		for (i = 8; i; i--) {
		    *output++ = limit[*output];
		    *output++ = limit[*output];
		    *output++ = limit[*output];
		    *output++ = limit[*output];
		    *output++ = limit[*output];
		    *output++ = limit[*output];
		    *output++ = limit[*output];
		    *output++ = limit[*output];

		    output += stride - 8;
		}
#else
		int value;
		seek_h262_coefficient_type block;
		unsigned *output_x20;
		unsigned char *output;

		output = blocks->base_address + offset;

		block = *blocks->coefficients;
		*blocks->coefficients = 0;
		if (block < 0) {
		    value = 256 - ((-(block | ~2047) + 4) >> 3);
		} else {
		    value = 256 + (((block & 2047) + 4) >> 3);
		}
		{
		    if (value < 256) {
			unsigned value_x7f7f7f7f;

			value |= value << 010;
			value |= value << 020;
			value_x7f7f7f7f = (unsigned) value >> 1;
			if (value & 1) {
			    value_x7f7f7f7f &= 0x7f7f7f7f;
			    for (value = 8; value; value--) {
				unsigned sum_0, sum_1, sum_2;

				output_x20 = (unsigned *) output;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_0 += sum_1;
				sum_1 ^= 0x01010101;
				sum_2 = sum_0 & 0x80808080;
				sum_1 &= sum_2 >> 7;
				sum_1 += (sum_0 & (sum_2 - (sum_2 >> 7))) << 1;
				*output_x20++ = sum_1;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_0 += sum_1;
				sum_1 ^= 0x01010101;
				sum_2 = sum_0 & 0x80808080;
				sum_1 &= sum_2 >> 7;
				sum_1 += (sum_0 & (sum_2 - (sum_2 >> 7))) << 1;
				*output_x20 = sum_1;
				output += stride;
			    }
			} else {
			    for (value = 8; value; value--) {
				unsigned sum_0, sum_1, sum_2;

				output_x20 = (unsigned *) output;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_2 = sum_0 & 0x80808080;
				sum_1 &= sum_2 >> 7;
				sum_1 += (sum_0 & (sum_2 - (sum_2 >> 7))) << 1;
				*output_x20++ = sum_1;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_2 = sum_0 & 0x80808080;
				sum_1 &= sum_2 >> 7;
				sum_1 += (sum_0 & (sum_2 - (sum_2 >> 7))) << 1;
				*output_x20 = sum_1;
				output += stride;
			    }
			}
		    } else {
			unsigned value_x7f7f7f7f;

			value &= 255;
			value |= value << 010;
			value |= value << 020;
			value_x7f7f7f7f = (unsigned) value >> 1;
			if (value & 1) {
			    value_x7f7f7f7f &= 0x7f7f7f7f;
			    for (value = 8; value; value--) {
				unsigned sum_0, sum_1;

				output_x20 = (unsigned *) output;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_0 += sum_1;
				sum_1 ^= 0x01010101;
				if (sum_0 & 0x80808080) {
				    unsigned mask;

				    mask = sum_0 & 0x80808080;
				    sum_0 ^= mask;
				    mask |= mask - (mask >> 7);
				    *output_x20++ =
					((sum_0 << 1) + sum_1) | mask;
				} else {
				    *output_x20++ = (sum_0 << 1) + sum_1;
				}
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				sum_0 += sum_1;
				sum_1 ^= 0x01010101;
				if (sum_0 & 0x80808080) {
				    unsigned mask;

				    mask = sum_0 & 0x80808080;
				    sum_0 ^= mask;
				    mask |= mask - (mask >> 7);
				    *output_x20 =
					((sum_0 << 1) + sum_1) | mask;
				} else {
				    *output_x20 = (sum_0 << 1) + sum_1;
				}
				output += stride;
			    }
			} else {
			    for (value = 8; value; value--) {
				unsigned sum_0, sum_1;

				output_x20 = (unsigned *) output;
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 &= 0x01010101;
				if (sum_0 & 0x80808080) {
				    unsigned mask;

				    mask = sum_0 & 0x80808080;
				    sum_0 ^= mask;
				    mask |= mask - (mask >> 7);
				    *output_x20++ =
					((sum_0 << 1) + sum_1) | mask;
				} else {
				    *output_x20++ = (sum_0 << 1) + sum_1;
				}
				sum_1 = *output_x20;
				sum_0 = ((sum_1 >> 1) & 0x7f7f7f7f)
				    + value_x7f7f7f7f;
				sum_1 = (sum_1 & 0x01010101);
				if (sum_0 & 0x80808080) {
				    unsigned mask;

				    mask = sum_0 & 0x80808080;
				    sum_0 ^= mask;
				    mask |= mask - (mask >> 7);
				    *output_x20 =
					((sum_0 << 1) + sum_1) | mask;
				} else {
				    *output_x20 = (sum_0 << 1) + sum_1;
				}
				output += stride;
			    }
			}
		    }
		}
#endif				/* 0 */
	    }
	}

	blocks++;
    }
}


void
mpeg3video_idct_conversion_1(struct seek_h262_block_type *blocks,
			     unsigned nblocks, unsigned offset,
			     unsigned stride)
{
    for (; nblocks; nblocks--) {
	{
	    {
		seek_h262_coefficient_type *lock;
		unsigned i = 8;
#if __ENABLE_BLOCK_META_DATA__
		unsigned meta_block;
#endif				/* __ENABLE_BLOCK_META_DATA__ */

		lock = blocks->coefficients;

#if __ENABLE_BLOCK_META_DATA__
# if __SHORT_ENDED_ARITHMETIC__
		meta_block = ((unsigned short *) lock)[-1];
# else
		meta_block = ((unsigned *) lock)[-1];
# endif				/* __SHORT_ENDED_ARITHMETIC__ */
		meta_block = ~meta_block << 1;
#endif				/* __ENABLE_BLOCK_META_DATA__ */

		for (; i; i--) {
		    int x0, x1, x2, x3, x4, x5, x6, x7, x8;

#if __ENABLE_BLOCK_META_DATA__
		    meta_block >>= 1;
		    if (meta_block & 1) {
			lock++;
			continue;
		    }
		    if (meta_block & 256) {
			lock[000] = lock[010] = lock[020] = lock[030] =
			    lock[040] = lock[050] = lock[060] = lock[070] =
			    lock[000] << 3;
			lock++;
			continue;
		    }

		    x0 = (lock[000] << 11) + 128;
		    x1 = lock[040] << 11;
		    x2 = lock[060];
		    x3 = lock[020];
		    x4 = lock[010];
		    x5 = lock[070];
		    x6 = lock[050];
		    x7 = lock[030];
#else
		    if (!((x1 = lock[040] << 11) | (x2 = lock[060])
			  | (x3 = lock[020]) | (x4 = lock[010])
			  | (x5 = lock[070]) | (x6 = lock[050])
			  | (x7 = lock[030]))) {
			lock[000] = lock[010] = lock[020] = lock[030] =
			    lock[040] = lock[050] = lock[060] = lock[070] =
			    lock[000] << 3;
			lock++;
			continue;
		    }

		    x0 = (lock[000] << 11) + 128;
#endif				/* __ENABLE_BLOCK_META_DATA__ */

		    x8 = W7 * (x4 + x5);
		    x4 = x8 + (W1 - W7) * x4;
		    x5 = x8 - (W1 + W7) * x5;
		    x8 = W3 * (x6 + x7);
		    x6 = x8 - (W3 - W5) * x6;
		    x7 = x8 - (W3 + W5) * x7;

		    x8 = x0 + x1;
		    x0 -= x1;
		    x1 = W6 * (x3 + x2);
		    x2 = x1 - (W2 + W6) * x2;
		    x3 = x1 + (W2 - W6) * x3;
		    x1 = x4 + x6;
		    x4 -= x6;
		    x6 = x5 + x7;
		    x5 -= x7;

		    x7 = x8 + x3;
		    x8 -= x3;
		    x3 = x0 + x2;
		    x0 -= x2;
		    x2 = (181 * (x4 + x5) + 128) >> 8;
		    x4 = (181 * (x4 - x5) + 128) >> 8;

		    lock[000] = (x7 + x1) >> 8;
		    lock[010] = (x3 + x2) >> 8;
		    lock[020] = (x0 + x4) >> 8;
		    lock[030] = (x8 + x6) >> 8;
		    lock[040] = (x8 - x6) >> 8;
		    lock[050] = (x0 - x4) >> 8;
		    lock[060] = (x3 - x2) >> 8;
		    lock[070] = (x7 - x1) >> 8;

		    lock++;
		}
	    }

	    {
		seek_h262_coefficient_type *lock;
		unsigned i = 8;
		unsigned char *output;

		output = blocks->base_address + offset;

		lock = blocks->coefficients;
		for (; i; i--) {
		    int x0, x1, x2, x3, x4, x5, x6, x7, x8;

#if __SHORT_ENDED_ARITHMETIC__
		    if (!lock[1]
			&& !(((int *) lock)[1] | ((int *) lock)[2]
			     | ((int *) lock)[3])) {
			unsigned value;

			value = ((lock[0] + 32) >> 6) & 0xff;
			value |= value << 010;
			value |= value << 020;

# if __GNUC__
			((unsigned *) output)[0] = value;
			((unsigned *) output)[1] = value;
# else
			{
			    unsigned *output_x20;

			    output_x20 = (unsigned *) output;
			    output_x20[0] = value;
			    output_x20[1] = value;
			}
# endif

			output += stride;

			DETACH_ROW(lock);

			continue;
		    }

		    x1 = lock[4] << 8;
		    x2 = lock[6];
		    x3 = lock[2];
		    x4 = lock[1];
		    x5 = lock[7];
		    x6 = lock[5];
		    x7 = lock[3];
#else
		    if (!((x1 = lock[4] << 8) | (x2 = lock[6]) | (x3 = lock[2])
			  | (x4 = lock[1]) | (x5 = lock[7]) | (x6 = lock[5])
			  | (x7 = lock[3]))) {
			unsigned value;

			value = ((lock[0] + 32) >> 6) & 0xff;
			value |= value << 010;
			value |= value << 020;

# if __GNUC__
			((unsigned *) output)[0] = value;
			((unsigned *) output)[1] = value;
# else
			{
			    unsigned *output_x20;

			    output_x20 = (unsigned *) output;
			    output_x20[0] = value;
			    output_x20[1] = value;
			}
# endif				/* __GNUC__ */

			output += stride;

			DETACH_ROW(lock);

			continue;
		    }
#endif				/* __SHORT_ENDED_ARITHMETIC__ */

		    x0 = (lock[0] << 8) + 8192 + (INTRA_IDCT_BIAS << 14);

		    x8 = W7 * (x4 + x5) + 4;
		    x4 = (x8 + (W1 - W7) * x4) >> 3;
		    x5 = (x8 - (W1 + W7) * x5) >> 3;
		    x8 = W3 * (x6 + x7) + 4;
		    x6 = (x8 - (W3 - W5) * x6) >> 3;
		    x7 = (x8 - (W3 + W5) * x7) >> 3;

		    x8 = x0 + x1;
		    x0 -= x1;
		    x1 = W6 * (x3 + x2) + 4;
		    x2 = (x1 - (W2 + W6) * x2) >> 3;
		    x3 = (x1 + (W2 - W6) * x3) >> 3;
		    x1 = x4 + x6;
		    x4 -= x6;
		    x6 = x5 + x7;
		    x5 -= x7;

		    x7 = x8 + x3;
		    x8 -= x3;
		    x3 = x0 + x2;
		    x0 -= x2;
		    x2 = (181 * (x4 + x5) + 128) >> 8;
		    x4 = (181 * (x4 - x5) + 128) >> 8;

#if __INTRA_CLIPPING__ == LOOKUP_CLIPPING
# if !WORDS_BIGENDIAN
		    x5 = range_limit[((x8 + x6) >> 016) & 2047];
		    x5 <<= 8;
		    x5 |= range_limit[((x0 + x4) >> 14) & 2047];
		    x5 <<= 8;
		    x5 |= range_limit[((x3 + x2) >> 14) & 2047];
		    x5 <<= 8;
		    x5 |= range_limit[((x7 + x1) >> 14) & 2047];
		    ((unsigned *) output)[0] = x5;
		    x5 = range_limit[((x7 - x1) >> 016) & 2047];
		    x5 <<= 8;
		    x5 |= range_limit[((x3 - x2) >> 14) & 2047];
		    x5 <<= 8;
		    x5 |= range_limit[((x0 - x4) >> 14) & 2047];
		    x5 <<= 8;
		    x5 |= range_limit[((x8 - x6) >> 14) & 2047];
		    ((unsigned *) output)[1] = x5;

		    output += stride;
# else
		    *output++ = range_limit[((x7 + x1) >> 14) & 2047];
		    *output++ = range_limit[((x3 + x2) >> 14) & 2047];
		    *output++ = range_limit[((x0 + x4) >> 14) & 2047];
		    *output++ = range_limit[((x8 + x6) >> 14) & 2047];
		    *output++ = range_limit[((x8 - x6) >> 14) & 2047];
		    *output++ = range_limit[((x0 - x4) >> 14) & 2047];
		    *output++ = range_limit[((x3 - x2) >> 14) & 2047];
		    *output++ = range_limit[((x7 - x1) >> 14) & 2047];

		    output += stride - 8;
# endif				/* !WORDS_BIGENDIAN */
#else
# if WORDS_BIGENDIAN
#  define H_SHIFT			020
#  define L_SHIFT			000
#  define SHIFT_E(c)			((c) << 010)
#  define SHIFT_O(c)			((c) << 000)
# else
#  define H_SHIFT			000
#  define L_SHIFT			020
#  define SHIFT_E(c)			((c) << 000)
#  define SHIFT_O(c)			((c) << 010)
# endif				/* WORDS_BIGENDIAN */
		    {
			unsigned k1, l2;

			SET(k1, ((x7 + x1) >> 14) << H_SHIFT);
			ADD(k1, ((x0 + x4) >> 14) << L_SHIFT);
			x5 = k1 & 0x02000200;
			k1 = k1 | (x5 - (x5 >> 9));
			x5 = k1 & 0x01000100;
			k1 = k1 & (x5 - (x5 >> 8));

			SET(l2, ((x3 + x2) >> 14) << H_SHIFT);
			ADD(l2, ((x8 + x6) >> 14) << L_SHIFT);
			x5 = l2 & 0x02000200;
			l2 = l2 | (x5 - (x5 >> 9));
			x5 = l2 & 0x01000100;
			l2 = l2 & (x5 - (x5 >> 8));

			((unsigned *) output)[0] = SHIFT_E(k1) | SHIFT_O(l2);

			SET(k1, ((x8 - x6) >> 14) << H_SHIFT);
			ADD(k1, ((x3 - x2) >> 14) << L_SHIFT);
			x5 = k1 & 0x02000200;
			k1 = k1 | (x5 - (x5 >> 9));
			x5 = k1 & 0x01000100;
			k1 = k1 & (x5 - (x5 >> 8));

			SET(l2, ((x0 - x4) >> 14) << H_SHIFT);
			ADD(l2, ((x7 - x1) >> 14) << L_SHIFT);
			x5 = l2 & 0x02000200;
			l2 = l2 | (x5 - (x5 >> 9));
			x5 = l2 & 0x01000100;
			l2 = l2 & (x5 - (x5 >> 8));

			((unsigned *) output)[1] = SHIFT_E(k1) | SHIFT_O(l2);
		    }

		    output += stride;
# if 1
#  undef H_SHIFT
#  undef L_SHIFT
#  undef SHIFT_E
#  undef SHIFT_O
# endif				/* WORDS_BIGENDIAN */
#endif				/* __INTRA_CLIPPING__ == LOOKUP_CLIPPING */

		    DETACH_ROW(lock);
		}
	    }
	}

	blocks++;
    }
}


void
mpeg3video_idct_conversion_3(void *video)
{
#if CLIP_LOOKUP
    if (range_limit_init) {
	limit_type *limit;
	unsigned i = LOOKUP_BIAS;

	limit = range_limit;

	for (; i; i--) {
	    *limit++ = 0;
	}
	for (; i < 0x100; i++) {
	    *limit++ = i;
	}
	for (i = 0x700 - LOOKUP_BIAS; i; i--) {
	    *limit++ = 255;
	}

	range_limit_init = 0;
    }
#endif				/* CLIP_LOOKUP */
}
