/*
 * e1f4.c
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008  Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <math.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include <e1f4.h>

#define __INTERFACE_SELECTOR__		0

#if !defined WORDS_BIGENDIAN
# define WORDS_BIGENDIAN 0
#endif				/* !WORDS_BIGENDIAN */

#define INTEGER_OUTPUT			0
#define REAL_OUTPUT			1

#define AUDIO_FAILURE			1
#define MALLOC_FAILURE			2
#define OPEN_FAILURE			3
#define READ_FAILURE			4
#define UNEXPECTED_EOF			6
#define WRITE_FAILURE			7

#define MMAP_SIZE			1048576
#define MMAP_SIZE_2			524288

#define DOUBLE_SYNCWORD_SYNCHRONIZATION	0
#define SELECT_SYNCWORD_SYNCHRONIZATION	1
#define SINGLE_SYNCWORD_SYNCHRONIZATION	2

#define ERROR(whatever) \
    (((struct Mylene1f4_whatever_type *) whatever)->error)
#define PRIVATE(whatever) \
    (((struct Mylene1f4_whatever_type *) whatever)->private)
#define SIZE(whatever) \
    (((struct Mylene1f4_whatever_type *) whatever)->size)
#define SYNCWORD_OFFSET(whatever) \
    (((struct Mylene1f4_whatever_type *) whatever)->syncword_offset)

#define PROGRESS_STAT \
    "\rframe, second, rate: %7u, %5u, %9.2f"
#define TRAILER_PRINT \
    "\rframes, seconds, rate, EOF: %7u, %5u, %9.2f, %+11d\n"

typedef struct extension_type {
    char *resource;
} extension_type;

typedef struct name_type {
    char *name;
    unsigned code, size, uniquem1;
} name_type;

extern Mylene1f4_real III_break_0[], III_break_1[];

char *file_name;
unsigned file_size;

static char *here_basename(const char *);

static int find_name(const struct name_type *, const char *, int *);
static int check_file(char *, unsigned *);

static void correct_underflow_0(void *);
static void do_release_0(void *, struct Mylene1f4_release_type *);
static void do_release_1(void *, struct Mylene1f4_release_type *);
static void do_release_3(void *, struct Mylene1f4_release_type *);
static void error_exit_2(void *);
static void header_hack_0(void *);
static void header_hack_1(struct Mylene1f4_header_type *);
static void init_audio(void);
static void init_tables(void);
static void open_qpcm(void);
static void print_trailer(void);
static void stat_audio(char *, void *, char *, unsigned);
static void stat_failure(char *, int);
static void stat_progress(void);
static void usage(void);

static Mylene1f4_real polyphase_buffer[1088], transform_buffer[2304];
static const char *layer_names[] = {
/* *INDENT-OFF* */
    NULL, "III", "II", "I"
/* *INDENT-ON* */
};
static char *prefix = NULL;
static int audio_initialized = 0, do_pcm = 0, do_explain = 0, fast_header, fd,
    fd2 = -1;
static jmp_buf env;
static Mylene1f4_audio_type e1f4_audio;
static const struct name_type output_names[] = {
/* *INDENT-OFF* */
    {	"integer",	INTEGER_OUTPUT,			0x8,	0	},
    {	"real",		REAL_OUTPUT,			0x5,	0	},
    {	NULL,		0,				0x0,	0	}
/* *INDENT-ON* */
}, process_names[] = {
/* *INDENT-OFF* */
    {	"4294967291",	Mylene1f4_4294967291,		0xb,	0	},
    {	"futile",	Mylene1f4_FUTILE,		0x7,	0	},
    {	"null",		Mylene1f4_FUTILE,		0x5,	0	},
    {	"pcm1",		Mylene1f4_4294967291,		0x5,	0	},
    {	"real",		Mylene1f4_4294967291,		0x5,	0	},
    {	"skip",		Mylene1f4_UTTERLY_FUTILE,	0x5,	0	},
    {	"utterly-futile",
			Mylene1f4_UTTERLY_FUTILE,	0xf,	0	},
    {	NULL,		0,				0x0,	0	}
/* *INDENT-ON* */
}, synchronization_methods[] = {
/* *INDENT-OFF* */
    {	"double",	DOUBLE_SYNCWORD_SYNCHRONIZATION,
							0x7,	0	},
    {	"select",	SELECT_SYNCWORD_SYNCHRONIZATION,
							0x7,	1	},
    {	"single",	SINGLE_SYNCWORD_SYNCHRONIZATION,
							0x7,	1	},
    {	NULL,		0,				0x0,	0	}
/* *INDENT-ON* */
};
static struct timeval start_time;
static unsigned header, mmap_offset, mmap_size, output_index,
    temporal_reference = 0;
static Mylene1f4_real output[2304];
static void *audio_buffer = NULL, *mmap_buffer;

static char *
here_basename(const char *path)
{
    char *ing;

    ing = (ing = strrchr(path, '/')) ? ing + 1 : (char *) path;

    return ing;
}


static int
check_file(char *file_name, unsigned *file_size)
{
    struct stat info;

    if (stat(file_name, &info)) {
	return -1;
    }
    if (!S_ISREG(info.st_mode)) {
	if (S_ISDIR(info.st_mode)) {
	    errno = EISDIR;
	} else {
	    errno = EINVAL;
	}

	return -1;
    }
    if (access(file_name, R_OK)) {
	return -1;
    }

    *file_size = info.st_size;

    return 0;
}


static int
find_name(const struct name_type *names, const char *string, int *code)
{
    unsigned length;

    length = strlen(string);
    for (; names->name; names++) {
	if (length < names->size
	    && (names->uniquem1 < length
		|| (names->uniquem1 == length && !names->name[length]))
	    && !memcmp(names->name, string, length)) {
	    *code = names->code;

	    return 1;
	}
    }

    return 0;
}


static void
correct_underflow_0(void *e1f4_whatever)
{
    struct Mylene1f4_whatever_type *whatever;
    unsigned file_offset = 0;

    whatever = (struct Mylene1f4_whatever_type *) e1f4_whatever;
    if (mmap_size) {
	file_offset = mmap_offset + mmap_size;
	munmap(mmap_buffer, mmap_size);
	mmap_offset += MMAP_SIZE_2;
    }
    mmap_size = MMAP_SIZE;
    if (file_size < mmap_offset + mmap_size) {
	mmap_size = file_size - mmap_offset;
    }
    if (whatever->size + mmap_offset + mmap_size - file_offset
	< whatever->minimum_useful_size) {
	longjmp(env, UNEXPECTED_EOF);
    }
    if ((mmap_buffer =
	 mmap(NULL, mmap_size, PROT_READ, MAP_SHARED, fd, mmap_offset))
	== MAP_FAILED) {
	mmap_size = 0;
	longjmp(env, READ_FAILURE);
    }
    whatever->size += mmap_offset + mmap_size - file_offset;
    whatever->buffer = (unsigned char *) mmap_buffer
	+ (mmap_size - whatever->size);
}


static void
do_release_0(void *e1f4_whatever, struct Mylene1f4_release_type *release)
{
}


static void
do_release_1(void *e1f4_whatever, struct Mylene1f4_release_type *release)
{
    if (fd2 == -1) {
	open_qpcm();
    }
    {
	unsigned size;

	size = release->flags & 32767;
	if (write(fd2, release->samples, size) != size) {
	    longjmp(env, WRITE_FAILURE);
	}
    }
}


static void
do_release_3(void *e1f4_whatever, struct Mylene1f4_release_type *release)
{
    if (fd2 == -1) {
	open_qpcm();
    }
    {
	Mylene1f4_real *source;
	short *target;
	unsigned i, size;

	source = (void *) release->samples;
	target = (void *) source;

	size = release->flags & 32767;

	for (i = size / sizeof(Mylene1f4_real); i; i--) {
	    Mylene1f4_real sample;

	    sample = *source++;
	    if ((Mylene1f4_real) 32767 < sample) {
		*target++ = +0x7fff;
	    } else if (sample < (Mylene1f4_real) -32768) {
		*target++ = -0x8000;
	    } else {
		*target++ = sample;
	    }
	}

	size >>= 1;

	if (write(fd2, release->samples, size) != size) {
	    longjmp(env, WRITE_FAILURE);
	}
    }
}


static void
error_exit_2(void *bull_0)
{
    switch (ERROR(e1f4_audio)) {
    case Mylene1f4_AUDIO_INVALID_BACKWARD_SEEK:
	/*
	 * should probably reset output
	 */
	stat_failure(((struct extension_type *) PRIVATE(e1f4_audio))->resource,
		     AUDIO_FAILURE);
	break;
    default:
	longjmp(env, AUDIO_FAILURE);
    }
}


static void
header_hack_0(void *e1f4_whatever)
{
    struct Mylene1f4_whatever_type *whatever;

    whatever = (struct Mylene1f4_whatever_type *) e1f4_whatever;

    if (do_explain) {
	if (temporal_reference) {
	    if (!(temporal_reference & 63)) {
		stat_progress();
	    }
	}
    }

    fast_header = 1;

    if (whatever->size - whatever->syncword_offset < 4) {
	whatever->minimum_useful_size = 4 + whatever->syncword_offset;
	Mylene1f4_correct_underflow(e1f4_whatever);
    }
    {
	unsigned new_header;

	new_header = *(unsigned *)
	    (whatever->buffer + whatever->syncword_offset);
#if WORDS_BIGENDIAN
	if ((new_header & 0xfff00000) != 0xfff00000
	    && 1 < temporal_reference) {
	    longjmp(env, ~0);
	}
#else
	if ((new_header & 0x0000f0ff) != 0x0000f0ff
	    && 1 < temporal_reference) {
	    longjmp(env, ~0);
	}
#endif				/* WORDS_BIGENDIAN */

	temporal_reference++;

	if (1 < do_explain) {
	    struct Mylene1f4_audio_attributes_type attributes;

#if WORDS_BIGENDIAN
	    attributes.header_hack_1 =
		(new_header & 0x0006fcc0) != (header & 0x0006fcc0)
		? header_hack_1 : NULL;
#else
	    attributes.header_hack_1 =
		(new_header & 0xc0fc0600) != (header & 0xc0fc0600)
		? header_hack_1 : NULL;
#endif				/* WORDS_BIGENDIAN */
	    Mylene1f4_set_audio_attributes(e1f4_audio, &attributes, 0,
					   Mylene1f4_AUDIO_HEADER_HACK_1_MASK);
	    header = new_header;
	}
    }

#if 0
    if ((*whatever->buffer != 255 || (whatever->buffer[1] & 240) != 240)
	&& 1 < temporal_reference) {
	longjmp(env, ~0);
    }
#endif				/* 0 */

    fast_header = 0;
}


static void
header_hack_1(struct Mylene1f4_header_type *header)
{
    if (temporal_reference != 1) {
	putchar('\n');
    }
    printf("layer %s", layer_names[header->layer]);
    {
	unsigned bitrate;

	Mylene1f4_bitrate(header, &bitrate);
	if (bitrate == Mylene1f4_FORBIDDEN) {
	    fputs(", forbidden bit rate", stdout);
	} else if (bitrate == Mylene1f4_FREE_FORMAT) {
	    fputs(", free format bit rate", stdout);
	} else {
	    printf(", %u kbits/s", bitrate);
	}
    }
    {
	unsigned sampling_frequency;

	Mylene1f4_sampling_frequency(header, &sampling_frequency);
	if (sampling_frequency == Mylene1f4_RESERVED) {
	    fputs(", reserved sampling frequency", stdout);
	} else {
	    printf(", %u Hz", sampling_frequency);
	}
    }
    putchar(' ');
    switch (header->mode) {
    case Mylene1f4_dual_channel:
	fputs("dual_channel", stdout);
	break;
    case Mylene1f4_joint_stereo:
	fputs("joint_stereo", stdout);
	break;
    case Mylene1f4_single_channel:
	fputs("single_channel", stdout);
	break;
    case Mylene1f4_stereo:
	fputs("stereo", stdout);
    }
    putchar('\n');
}


static void
init_audio(void)
{
    Mylene1f4_init_tables();
    init_tables();
    audio_initialized = 1;
}


static void
init_tables(void)
{
    int i = 0;

    for (; i < 8207; i++) {
	III_break_0[i] = pow(i, (double) 4 / 3);
    }
    for (i = -256; i < 122; i++) {
	III_break_1[i + 256] =
	    pow((double) 2.0, -0.25 * (double) (i + 210));
    }
}


static void
open_qpcm(void)
{
    char *base, *path;

    base = here_basename(file_name);
    if (!prefix) {
	path = base;
    } else {
	unsigned length, size;

	length = strlen(prefix);
	size = strlen(base);
	path = (char *) malloc(length + size + 1);
	if (!path) {
	    longjmp(env, MALLOC_FAILURE);
	}
	memcpy(path, prefix, length);
	memcpy(path + length, base, size + 1);
    }
    fd2 = open(path, O_CREAT | O_TRUNC | O_WRONLY, 00600);
    if (prefix) {
	free(path);
    }
    if (fd2 == -1) {
	longjmp(env, OPEN_FAILURE);
    }
}


static void
print_trailer(void)
{
    double time;
    int bytes_to_EOF;
    struct timeval current_time;

    bytes_to_EOF = file_size - (mmap_offset + mmap_size);
    bytes_to_EOF += SIZE(e1f4_audio) - SYNCWORD_OFFSET(e1f4_audio);
    gettimeofday(&current_time, NULL);
    time = current_time.tv_sec - start_time.tv_sec;
    time += (double) (current_time.tv_usec - start_time.tv_usec) / 1000000;
    fprintf(stderr, TRAILER_PRINT, temporal_reference, (unsigned) time,
	    (double) temporal_reference / time, bytes_to_EOF);
}


static void
stat_audio(char *argv_0, void *e1f4_audio, char *file_name,
	   unsigned file_offset)
{
    char *error_string;

    Mylene1f4_audio_error_string(ERROR(e1f4_audio), &error_string);
    fprintf(stderr, "%s: audio parse error in `%s' at %08x (%d) (%s)\n",
	    argv_0, file_name, file_offset, file_offset, error_string);
}


static void
stat_failure(char *me, int error)
{
    if (do_explain && 1 < temporal_reference) {
	fputc('\n', stdout);
	fflush(stdout);
    }

    switch (error) {
    case AUDIO_FAILURE:
	stat_audio(me, e1f4_audio, file_name,
		   mmap_offset + mmap_size - SIZE(e1f4_audio));
	break;
    case MALLOC_FAILURE:
	fprintf(stderr, "%s: fail to allocate memory for `%s'\n", me,
		file_name);
	break;
    case OPEN_FAILURE:
	fprintf(stderr, "%s: cannot open `%s%s': %s\n", me,
		prefix ? prefix : "", here_basename(file_name),
		strerror(errno));
	break;
    case READ_FAILURE:
	fprintf(stderr, "%s: cannot read `%s'\n", me, file_name);
	break;
    case UNEXPECTED_EOF:
	fprintf(stderr, "%s: unexpected EOF in `%s' at %08x\n", me, file_name,
		mmap_offset + mmap_size/* - SIZE(e1f4_audio)*/);
	break;
    case WRITE_FAILURE:
	fprintf(stderr, "%s: cannot write `%s%s': %s\n", me,
		prefix ? prefix : "", here_basename(file_name),
		strerror(errno));
    }
}


static void
stat_progress(void)
{
    double time;
    struct timeval current_time;

    gettimeofday(&current_time, NULL);
    time = current_time.tv_sec - start_time.tv_sec;
    time += (double) (current_time.tv_usec - start_time.tv_usec) / 1000000;
    printf(PROGRESS_STAT, temporal_reference, (unsigned) time,
	    (double) temporal_reference / time);
    fflush(stdout);
}


static void
usage(void)
{
    puts("Usage: ef14 [OPTIONS] FILE\n\
Decode audio streams in FILE(s).\n\
\n\
Options:\n\
  -P, --process 4294967291|pcm1|real\n\
				decode audio stream (default)\n\
  -P, --process futile|null	parse audio stream without actually decoding\
 it\n\
  -P, --process skip|utterly-futile\n\
				skip audio data parsing and decoding all\n\
				together\n\
  -m, --synchronization-method double|select|single\n\
				set synchronization method\n\
  -o, --output integer|real	set decoder output type\n\
  -p, --pcm PREFIX		?\n\
  -s, --frames-count COUNT	set frames count for `select' synchronization\
\n\
				method (defaults to 4)\n\
  -v, --verbose			?\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int output_type = INTEGER_OUTPUT, process = Mylene1f4_4294967291,
	synchronization = DOUBLE_SYNCWORD_SYNCHRONIZATION, status = 0;
    struct extension_type extension_data;
    unsigned nfiles, select = 4;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *frames_count_specs = NULL, *output_specs = NULL,
	    *process_specs = NULL, *synchronization_specs = NULL;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "frames-count", 0x01,   NULL,   0x73   },
		{   "output",	    0x01,   NULL,   0x6f   },
		{   "pcm",	    0x01,   NULL,   0x70   },
		{   "process",	    0x01,   NULL,   0x50   },
		{   "synchronization-method",
				    0x01,   NULL,   0x6d   },
		{   "verbose",	    0x00,   NULL,   0x76   },
		{   "version",	    0x00,   NULL,   0x56   },
		{   "help",	    0x00,   NULL,   0x48   },
		{   NULL,	    0x00,   NULL,   0x00   }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "P:m:o:p:s:v", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'P':
		process_specs = optarg;
		break;
	    case 'm':
		synchronization_specs = optarg;
		break;
	    case 'o':
		output_specs = optarg;
		break;
	    case 'p':
		do_pcm = 1;
		prefix = optarg;
		break;
	    case 's':
		frames_count_specs = optarg;
		break;
	    case 'v':
		do_explain++;
		break;
	    case 'H':
		usage();

		return 0;
	    case 'V':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);

		return 0;
	    case '?':
		return 1;
	    }
	}

	if (output_specs) {
	    if (!find_name(output_names, output_specs, &output_type)) {
		fprintf(stderr, "%s: cannot parse decoder output type specif"
			"ication: `%s'\n", argv[0], output_specs);

		return 1;
	    }
	}

	if (process_specs) {
	    if (!find_name(process_names, process_specs, &process)) {
		fprintf(stderr, "%s: cannot parse process specification: `%s"
			"'\n", argv[0], process_specs);

		return 1;
	    }
	}

	if (synchronization_specs) {
	    if (!find_name(synchronization_methods, synchronization_specs,
			   &synchronization)) {
		fprintf(stderr, "%s: cannot parse synchronization method spe"
			"cification: `%s'\n", argv[0], synchronization_specs);

		return 1;
	    }
	}

	if (synchronization == SELECT_SYNCWORD_SYNCHRONIZATION) {
	    if (frames_count_specs) {
		char *error;

		select = strtoul(frames_count_specs, &error, 0);
		if (error == frames_count_specs || *error) {
		    fprintf(stderr, "%s: cannot parse frame count specificat"
			    "ion for `select' synchronization method: `%s'\n",
			    argv[0], frames_count_specs);

		    return 1;
		}
	    }
	}
    }

    if (argc < 2) {
	fprintf(stderr, "%s: too few arguments\nTry `%s --help' for more inf"
		"ormation\n", argv[0], argv[0]);

	return 1;
    }

    extension_data.resource = argv[0];

    nfiles = argc - optind;
    for (; optind < argc; optind++) {
	volatile unsigned success = 0;

	file_name = argv[optind];
	if (do_explain && nfiles != 1) {
	    printf("%s:\n", file_name);
	}
	do {
	    if (check_file(file_name, &file_size)) {
		fprintf(stderr, "%s: fail to load `%s'\n", argv[0], file_name);
		break;
	    }
	    if ((fd = open(file_name, O_RDONLY)) == -1) {
		fprintf(stderr, "%s: fail to open `%s'\n", argv[0], file_name);
		break;
	    }
	    success++;
	    {
		int error;
		unsigned size;

		Mylene1f4_audio_size(&size);
		{
		    if (!(e1f4_audio = (Mylene1f4_audio_type)
			  malloc(size))) {
			fprintf(stderr, "%s: fail to allocate memory for `%s' "
				"audio\n", argv[0], file_name);
			break;
		    }
		}
		Mylene1f4_reset_audio(e1f4_audio);

		{
		    struct Mylene1f4_audio_attributes_type attributes;
		    unsigned extra_mask = 0, mask = 0;

		    attributes.abort_decoding = error_exit_2;
		    attributes.buffer = NULL;
		    attributes.correct_underflow = correct_underflow_0;
		    attributes.header_hack_0 = header_hack_0;
		    attributes.output = (void *) output;
		    attributes.polyphase_buffer = polyphase_buffer;
		    attributes.process = process;
		    attributes.size = 0;
		    attributes.transform_buffer = transform_buffer;

		    if (output_type == INTEGER_OUTPUT) {
			attributes.do_release =
			    do_pcm ? do_release_1 : do_release_0;
		    } else {
			attributes.do_release =
			    do_pcm ? do_release_3 : do_release_0;
			attributes.polyphase_synthesis_0 =
			    Mylene1f4_polyphase_synthesis_8;
			attributes.polyphase_synthesis_1 =
			    Mylene1f4_polyphase_synthesis_9;
			attributes.sample_shift = 2;
			mask |= Mylene1f4_AUDIO_POLYPHASE_SYNTHESIS_0_MASK
			    | Mylene1f4_AUDIO_POLYPHASE_SYNTHESIS_1_MASK
			    | Mylene1f4_AUDIO_SAMPLE_SHIFT_MASK;
		    }

		    Mylene1f4_set_audio_attributes
			(e1f4_audio, &attributes,
			 Mylene1f4_AUDIO_ABORT_DECODING_MASK
			 | Mylene1f4_AUDIO_BUFFER_MASK
			 | Mylene1f4_AUDIO_CORRECT_UNDERFLOW_MASK
			 | Mylene1f4_AUDIO_DO_RELEASE_MASK
			 | Mylene1f4_AUDIO_OUTPUT_MASK
			 | Mylene1f4_AUDIO_POLYPHASE_BUFFER_MASK
			 | Mylene1f4_AUDIO_PROCESS_MASK
			 | Mylene1f4_AUDIO_SIZE_MASK
			 | Mylene1f4_AUDIO_TRANSFORM_BUFFER_MASK | mask, 
			 Mylene1f4_AUDIO_HEADER_HACK_0_MASK | extra_mask);
		}

		PRIVATE(e1f4_audio) = &extension_data;

		if ((error = setjmp(env))) {
		    if (fast_header && error != READ_FAILURE) {
			error = 0;
		    } else {
			stat_failure(argv[0], error);
		    }
		} else {
		    if (!audio_initialized) {
			init_audio();
		    }
		    fast_header = 0;
		    header = 0;
		    mmap_offset = 0;
		    mmap_size = 0;
		    output_index = 0;
		    if (do_explain) {
			temporal_reference = 0;
			gettimeofday(&start_time, NULL);
		    }
		    if (synchronization == SINGLE_SYNCWORD_SYNCHRONIZATION) {
			Mylene1f4_find_audio(e1f4_audio);
		    } else if (synchronization
			       == DOUBLE_SYNCWORD_SYNCHRONIZATION) {
			Mylene1f4_seek_audio(e1f4_audio);
		    } else {
			Mylene1f4_synchronize_audio(e1f4_audio, select);
		    }
#if !__INTERFACE_SELECTOR__
		    Mylene1f4_audio_sequence(e1f4_audio);
#else
#endif				/* !__INTERFACE_SELECTOR__ */
		}

		if (do_explain && !error) {
		    print_trailer();
		}

		{
		    if (fd2 != -1) {
			close(fd2);
			fd2 = -1;
		    }
		}

		if (audio_buffer) {
		    free(audio_buffer);
		}

		if (mmap_size) {
		    munmap(mmap_buffer, mmap_size);
		}

		free(e1f4_audio);
	    }
	} while (0);
	if (0 < success) {
	    close(fd);
	}
    }

    return status;
}
