/*
 * h262.c
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008  Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <setjmp.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include <../dct/idct.0.h>
#include <../dct/idct.1.h>
#include <h262.h>

#define __INTERFACE_SELECTOR__		1

#if !defined WORDS_BIGENDIAN
# define WORDS_BIGENDIAN 0
#endif				/* !WORDS_BIGENDIAN */

#if WORDS_BIGENDIAN
# define MPEG_program_end_code		0x000001b9
# define iso_11172_end_code		0x000001b9
# define pack_start_code		0x000001ba
# define sequence_end_code		0x000001b7
# define sequence_header_code		0x000001b3
# define system_header_start_code	0x000001bb
#else
# define MPEG_program_end_code		0xb9010000
# define iso_11172_end_code		0xb9010000
# define pack_start_code		0xba010000
# define sequence_end_code		0xb7010000
# define sequence_header_code		0xb3010000
# define system_header_start_code	0xbb010000
#endif				/* WORDS_BIGENDIAN */

#define ALLOCATION_FAILURE		0x1
#define IDENTIFICATION_FAILURE		0x2
#define MALLOC_FAILURE			0x3
#define PACKET_START_CODE_MISMATCH	0x4
#define READ_FAILURE			0x5
#define SEEK_FAILURE			0x6
#define SEQUENCE_SCALABLE_EXTENSION_SEEN \
					0x7
#define SYSTEM_FAILURE			0x8
#define UNEXPECTED_EOF			0x9
#define UNEXPECTED_START_CODE		0xa
#define VIDEO_FAILURE			0xb

#define BUFFER(whatever) \
    (((struct seek_h262_whatever_type *) whatever)->buffer)
#define ERROR(whatever) \
    (((struct seek_h262_whatever_type *) whatever)->error)
#define SIZE(whatever) \
    (((struct seek_h262_whatever_type *) whatever)->size)
#define START_CODE(whatever) \
    (((struct seek_h262_whatever_type *) whatever)->start_code)

#define STORAGE_SIZE			4096
#define VIDEO_BUFFER_SIZE \
    ((65536 + ((SEEK_H262_MAXIMUM_MACROBLOCK_SIZE + 7) >> 3) + 3) & ~3)

#if WORDS_BIGENDIAN
# define start_code(start_code) ((start_code) & 255)
# define stream_id(start_code) ((start_code) & 0xff)
#else
# define start_code(start_code) ((start_code) >> 24)
# define stream_id(start_code) ((start_code) >> 030)
#endif				/* WORDS_BIGENDIAN */

typedef struct name_type {
    char *name;
    unsigned code, size, uniquem1;
} name_type;

static char *here_basename(const char *);

static int decode_stream(unsigned, unsigned, unsigned);
static int decode_system(unsigned);
static int find_name(const struct name_type *, char *, int *);
static int parse_set(unsigned char *, unsigned, unsigned, unsigned char *);
static int read_buffer(void *, unsigned);
static int read_hexadecimal(unsigned char *, unsigned *);
static int select_system(unsigned *, unsigned *, unsigned *);

static unsigned file_offset(void);

static void alloc_storage(void *, unsigned char **, unsigned, unsigned);
static void correct_underflow_0(void *);
static void correct_underflow_2(void *);
static void correct_underflow_3(void *);
static void do_release_0(void *, struct seek_h262_release_type *);
static void do_release_1(void *, struct seek_h262_release_type *);
static void error_exit_0(void *);
static void error_exit_2(void *);
static void packet_data(void *, unsigned);
static void read_packet(void *, unsigned);
static void seek_stream(void *, unsigned);
static void stat_failure(int);
static void stat_progress(void);
static void stat_system(char *, void *, char *, unsigned);
static void stat_video(char *, void *, char *, unsigned);
static void x00000100___hack_1_0(struct seek_h262_x00000100___type *);
static void x00000100___hack_1_2(struct seek_h262_x00000100___type *);
static void x00000100___hack_1_3(struct seek_h262_x00000100___type *);
static void x000001b5_1_hack_1_1(struct seek_h262_x000001b5_1_type *);
static void x000001b5_1_hack_1_2(struct seek_h262_x000001b5_1_type *);
static void x000001b5_5_hack_0(void *);
static void x000001ff___hack_2_1(struct seek_h262_x000001ff___type *);
static void usage(void);

static const char *picture_coding_type_names[] = {
/* *INDENT-OFF* */
    NULL,
    "intra_coded",
    "predictive_coded",
    "bidirectionally_predictive_coded",
    "dc_intra_coded"
/* *INDENT-ON* */
};

static char *argv_0, *file_name, *prefix = NULL, *name = NULL;
static int color_space = SEEK_H262_YCbCr, do_ppm = 0, do_progress = 0,
    do_time = 0, fd, have_start_code, is_system, new_line = 0,
    skip_bidirectionally_predictive_coded = 0, skip_predictive_coded = 0,
    swap_buffers, temporar_color_space, video_initialized = 0;
#if __Berkeley_FUCK_UP__ == 1
static int Berkeley_fuck_up = 0;
#endif				/* __Berkeley_FUCK_UP__ == 1 */
#if __INTERFACE_SELECTOR__
static int synthetic_sequence_extension;
#endif				/* __INTERFACE_SELECTOR__ */
static jmp_buf env;
static seek_h262_system_type h262_system;
static seek_h262_video_type h262_video;
static const struct {
    void (*idct_add_block) (struct seek_h262_block_type *, unsigned, unsigned,
			    unsigned),
	(*idct_set_block) (struct seek_h262_block_type *, unsigned, unsigned,
			   unsigned),
	(*init_idct) (void *);
} idcts[] = {
/* *INDENT-OFF* */
    {	dummy_idct_conversion_0,
	dummy_idct_conversion_1,
	dummy_idct_conversion_3					},
    {	mpeg3video_idct_conversion_0,
	mpeg3video_idct_conversion_1,
	mpeg3video_idct_conversion_3				}
/* *INDENT-ON* */
};
static const struct name_type color_spaces[] = {
/* *INDENT-OFF* */
    {	"GRAYSCALE",	SEEK_H262_GRAYSCALE,		0xa,	0	},
    {	"YCbCr",	SEEK_H262_YCbCr,		0x6,	0	},
    {	"none",		SEEK_H262_none,			0x5,	0	},
    {	NULL,		0,				0x0,	0	}
/* *INDENT-ON* */
}, idct_routines[] = {
/* *INDENT-OFF* */
    {	"dummy",			0,		006,	0x0	},
    {	"mpeg3video_idct_conversion",	1,		033,	0x0	},
    {	NULL,				0,		000,	0x0	}
/* *INDENT-ON* */
};
static struct timeval start_time;
static unsigned PES_packet_data_size, chroma_format, file_index,
    horizontal_size, nvideo_streams = 0, prefix_length = 0,
    stream_excess, stream_id, system_excess,
    system_offset, temporal_reference = 0, vertical_size;
static unsigned char *stream_storage, *system_storage, video_streams[16];
static void (*idct_add_method) (struct seek_h262_block_type *, unsigned,
				unsigned, unsigned),
    (*idct_set_method) (struct seek_h262_block_type *, unsigned, unsigned,
			unsigned),
    (*init_video_idct) (void *),
    (*pack_header) (seek_h262_system_type),
    (*packet) (seek_h262_system_type), *video_buffer = NULL;

static char *
here_basename(const char *path)
{
    char *ing;

    ing = (ing = strrchr(path, '/')) ? ing + 1 : (char *) path;

    return ing;
}


static int
decode_stream(unsigned file_offset, unsigned first_start_code,
	      unsigned system_type)
{
    int error = 0;

    if (first_start_code == sequence_header_code) {
	is_system = 0;
    } else {
	unsigned size;

	seek_h262_system_size(&size);
	if (!(h262_system = (seek_h262_system_type)
	      malloc(size + STORAGE_SIZE))) {
	    return MALLOC_FAILURE;
	}
	seek_h262_reset_system(h262_system);

	system_storage = (unsigned char *) h262_system + size;

	{
	    struct seek_h262_system_attributes_type attributes;

	    attributes.abort_decoding = error_exit_0;
	    attributes.correct_underflow = correct_underflow_2;
	    attributes.packet_data = packet_data;
	    attributes.seek_stream = seek_stream;
	    system_offset = STORAGE_SIZE;
	    attributes.x000001ff___hack_2 = x000001ff___hack_2_1;

	    seek_h262_set_system_attributes
		(h262_system, &attributes,
		 SEEK_H262_SYSTEM_ABORT_DECODING_MASK
		 | SEEK_H262_SYSTEM_CORRECT_UNDERFLOW_MASK
		 | SEEK_H262_SYSTEM_PACKET_DATA_MASK
		 | SEEK_H262_SYSTEM_SEEK_STREAM_MASK
		 | SEEK_H262_SYSTEM_X000001FF___HACK_2_MASK);
	}

	if (system_type == 11172) {
	    pack_header = seek_h262_artificial_pack_header;
	    packet = seek_h262_packet;
	} else {
	    pack_header = seek_h262_pack_header;
	    packet = seek_h262_PES_packet;
	}

	is_system = 3;
	system_storage = (unsigned char *) ((char *) h262_system + size);
    }

    for (stream_id = 0xe0; stream_id < 0xf0; stream_id++) {
	if (is_system && !video_streams[stream_id - 0xe0]) {
	    continue;
	}

	if (is_system) {
	    BUFFER(h262_system) = NULL;
	    SIZE(h262_system) = 0;

	    lseek(fd, file_offset + 4, SEEK_SET);

	    have_start_code = 1;

	    START_CODE(h262_system) = pack_start_code;
	}

	{
	    char *base_name = NULL;
	    unsigned base_name_length = 0, max_name_length = 0, size;

	    seek_h262_video_size(&size);

	    if (do_ppm) {
		base_name = here_basename(file_name);
		base_name_length = strlen(base_name);
		if (prefix) {
		    prefix_length = strlen(prefix);
		}
		max_name_length =
		    prefix_length + base_name_length + 50 + is_system;
	    }
	    if (!(h262_video = (seek_h262_video_type)
		  malloc(size + VIDEO_BUFFER_SIZE + max_name_length))) {
		error = MALLOC_FAILURE;
		break;
	    }
	    stream_storage = (unsigned char *) h262_video + size;
	    if (max_name_length) {
		name = (char *) stream_storage + VIDEO_BUFFER_SIZE;
		memcpy(name, prefix, prefix_length);
		memcpy(name + prefix_length, base_name, base_name_length);
		prefix_length += base_name_length;
	    }
	}

	seek_h262_reset_video(h262_video);

	{
	    struct seek_h262_video_attributes_type attributes;
	    unsigned extra_mask = 0, mask = 0;

#if __Berkeley_FUCK_UP__ == 1
	    attributes.Berkeley_fuck_up = Berkeley_fuck_up;
#endif				/* __Berkeley_FUCK_UP__ == 1 */
	    if (do_progress) {
		extra_mask = SEEK_H262_VIDEO_X00000100___HACK_1_MASK;
		attributes.x00000100___hack_1 = x00000100___hack_1_2;
	    } else if (skip_bidirectionally_predictive_coded
		       || skip_predictive_coded) {
		attributes.x00000100___hack_1 = x00000100___hack_1_0;
		extra_mask = SEEK_H262_VIDEO_X00000100___HACK_1_MASK;
	    } else if (do_ppm || do_time) {
		attributes.x00000100___hack_1 = x00000100___hack_1_3;
		extra_mask = SEEK_H262_VIDEO_X00000100___HACK_1_MASK;
	    }
	    if (do_ppm
		&& !(extra_mask
		     & SEEK_H262_VIDEO_X000001B5_1_HACK_1_MASK)) {
		extra_mask |= SEEK_H262_VIDEO_X000001B5_1_HACK_1_MASK;
		attributes.x000001b5_1_hack_1 = x000001b5_1_hack_1_1;
	    } else if (!(extra_mask
			 & SEEK_H262_VIDEO_X000001B5_1_HACK_1_MASK)) {
		extra_mask |= SEEK_H262_VIDEO_X000001B5_1_HACK_1_MASK;
		attributes.x000001b5_1_hack_1 = x000001b5_1_hack_1_2;
	    }
	    if (color_space != SEEK_H262_none) {
		attributes.idct_add_block = idct_add_method;
		attributes.idct_set_block = idct_set_method;
		mask |= SEEK_H262_VIDEO_IDCT_ADD_BLOCK_MASK;
		mask |= SEEK_H262_VIDEO_IDCT_SET_BLOCK_MASK;
	    }
	    if (do_ppm) {
		attributes.do_release = do_release_1;
	    } else {
		attributes.do_release = do_release_0;
	    }
	    attributes.abort_decoding = error_exit_2;
	    attributes.alloc_storage = alloc_storage;
	    attributes.buffer = NULL;
	    attributes.color_space = color_space;
	    if (is_system) {
		attributes.correct_underflow = correct_underflow_3;
	    } else {
		attributes.correct_underflow = correct_underflow_0;
	    }
	    attributes.size = 0;
	    attributes.x000001b5_5_hack_0 = x000001b5_5_hack_0;

	    seek_h262_set_video_attributes
		(h262_video, &attributes,
		 SEEK_H262_VIDEO_ABORT_DECODING_MASK
		 | SEEK_H262_VIDEO_ALLOC_STORAGE_MASK
		 | SEEK_H262_VIDEO_BUFFER_MASK
#if __Berkeley_FUCK_UP__ == 1
		 | SEEK_H262_VIDEO_Berkeley_FUCK_UP_MASK
#endif				/* __Berkeley_FUCK_UP__ == 1 */
		 | SEEK_H262_VIDEO_COLOR_SPACE_MASK
		 | SEEK_H262_VIDEO_CORRECT_UNDERFLOW_MASK
		 | SEEK_H262_VIDEO_DO_RELEASE_MASK
		 | SEEK_H262_VIDEO_SIZE_MASK | mask,
		 SEEK_H262_VIDEO_X000001B5_5_HACK_0_MASK
		 | extra_mask);

	    if (color_space != SEEK_H262_none) {
		init_video_idct(h262_video);
	    }

	    if (is_system) {
		swap_buffers = 0;
	    } else {
	    }

	    if (do_ppm) {
		file_index = ~0;
	    }
	}

	if ((error = setjmp(env))) {
	    if (do_progress) {
		if (temporal_reference) {
		    temporal_reference--;
		    stat_progress();
		    fputc('\n', stdout);
		    fflush(stdout);
		}
	    }
	} else {
	    if (!video_initialized) {
		video_initialized = 1;
		seek_h262_init_tables();
	    }
	    stream_excess = 0;
	    new_line = 0;
	    if (do_time) {
		temporal_reference = 0;
		gettimeofday(&start_time, NULL);
	    }
#if !__INTERFACE_SELECTOR__
	    seek_h262_video_sequence(h262_video);
#else
	    seek_h262_next_x00000100(h262_video);
	    if (synthetic_sequence_extension) {
		do {
		    seek_h262_next_11172_x00000100(h262_video);
		} while (START_CODE(h262_video) != sequence_end_code);
	    } else {
		do {
		    seek_h262_next_13818_x00000100(h262_video);
		} while (START_CODE(h262_video) != sequence_end_code);
	    }
	    seek_h262_do_release(h262_video);
#endif				/* !__INTERFACE_SELECTOR__ */

	    if (do_time) {
		double time;
		struct timeval current_time;

		gettimeofday(&current_time, NULL);
		time = current_time.tv_sec - start_time.tv_sec;
		time += (double)
		    (current_time.tv_usec - start_time.tv_usec)
		    / 1000000;
		printf("\rframes, seconds, rate: %7u, %5u, %9.2f\n",
		       temporal_reference, (unsigned) time,
		       (double) temporal_reference / time);
	    }
	}

	if (video_buffer) {
	    free(video_buffer);
	}

	if (error) {
	    stat_failure(error);
	    error = 0;
	}

	free(h262_video);

	if (!is_system) {
	    break;
	}
    }

    if (is_system) {
	free(h262_system);
    }

    return error;
}


static int
decode_system(unsigned system_type)
{
    int error = 0;
    unsigned size;

    if (do_progress) {
	if (lseek(fd, 4, SEEK_CUR) == -1) {
	    return SEEK_FAILURE;
	}
    }

    seek_h262_system_size(&size);
    if (!(h262_system = (seek_h262_system_type) malloc(size + STORAGE_SIZE))) {
	return MALLOC_FAILURE;
    }
    seek_h262_reset_system(h262_system);

    system_storage = (unsigned char *) h262_system + size;

    {
	struct seek_h262_system_attributes_type attributes;
	unsigned mask = 0;

	attributes.abort_decoding = error_exit_0;
	attributes.buffer = system_storage;
	attributes.correct_underflow = correct_underflow_2;
	attributes.packet_data = packet_data;
	attributes.seek_stream = seek_stream;
	attributes.size = 0;

	system_offset = STORAGE_SIZE;

	if (do_progress) {
	    START_CODE(h262_system) = pack_start_code;
	}

	seek_h262_set_system_attributes
	    (h262_system, &attributes,
	     SEEK_H262_SYSTEM_ABORT_DECODING_MASK
	     | SEEK_H262_SYSTEM_BUFFER_MASK
	     | SEEK_H262_SYSTEM_CORRECT_UNDERFLOW_MASK
	     | SEEK_H262_SYSTEM_PACKET_DATA_MASK
	     | SEEK_H262_SYSTEM_SEEK_STREAM_MASK
	     | SEEK_H262_SYSTEM_SIZE_MASK | mask);
    }

    if ((error = setjmp(env))) {
	if (do_progress) {
	    fputc('\n', stdout);
	    fflush(stdout);
	}
    } else if (system_type == 11172) {
	if (do_progress) {
	    while (1) {
		if (START_CODE(h262_system) == pack_start_code) {
			seek_h262_artificial_pack_header(h262_system);
		} else if (system_header_start_code
			   < START_CODE(h262_system)) {
		    unsigned file_offset;

		    seek_h262_packet(h262_system);
		    file_offset = lseek(fd, 0L, SEEK_CUR);
		    if ((int) file_offset == -1) {
			break;
		    }
		    printf("\r%u", file_offset);
		    fflush(stdout);
		    seek_h262_next_start_code(h262_system);
		} else {
		    break;
		}
	    }
	} else {
	    seek_h262_iso11172_stream(h262_system);
	}
    } else {
	if (do_progress) {
	    while (1) {
		if (START_CODE(h262_system) == pack_start_code) {
		    seek_h262_pack_header(h262_system);
		} else if (system_header_start_code
			   < START_CODE(h262_system)) {
		    unsigned file_offset;

		    seek_h262_PES_packet(h262_system);
		    file_offset = lseek(fd, 0L, SEEK_CUR);
		    if ((int) file_offset == -1) {
			break;
		    }
		    printf("\r%u", file_offset);
		    fflush(stdout);
		    seek_h262_next_start_code(h262_system);
		} else {
		    break;
		}
	    }
	} else {
	    seek_h262_MPEG2_program_stream(h262_system);
	}
    }

    if (error) {
	stat_failure(error);
	error = 0;
    } else {
	if (do_progress) {
	    if ((int) file_offset == -1) {
		error = SEEK_FAILURE;
	    } else if (START_CODE(h262_system) != iso_11172_end_code) {
		error = UNEXPECTED_START_CODE;
	    } else {
		putchar('\n');
	    }
	}
    }

    free(h262_system);

    return error;
}


static int
find_name(const struct name_type *names, char *string, int *code)
{
    unsigned length;

    length = strlen(string);
    for (; names->name; names++) {
	if (length < names->size
	    && (names->uniquem1 < length
		|| (names->uniquem1 == length && !names->name[length]))
	    && !memcmp(names->name, string, length)) {
	    *code = names->code;

	    return 1;
	}
    }

    return 0;
}


static int
parse_set(unsigned char *set, unsigned offset, unsigned size,
	  unsigned char *specification)
{
    int ing = 0;

    memset(set, '\0', size);
    if (specification) {
	unsigned length;

	length = strlen((char *) specification);
	while (length) {
	    unsigned item;

	    if (length < 2) {
		ing = -1;
		break;
	    }
	    if (read_hexadecimal(specification, &item)) {
		ing = -1;
		break;
	    }
	    length -= 2;
	    specification += 2;
	    if (length && *specification == 45) {
		unsigned second_item;

		if (length < 3) {
		    ing = -1;
		    break;
		}
		if (read_hexadecimal(specification + 1, &second_item)) {
		    ing = -1;
		    break;
		}
		length -= 3;
		specification += 3;
		if (second_item < item) {
		    unsigned swap;

		    swap = item;
		    item = second_item;
		    second_item = swap;
		}
		if (item < offset) {
		    item = offset;
		} else if (offset + size < item + 1) {
		    item = offset + size - 1;
		}
		if (second_item < offset) {
		    second_item = offset;
		} else if (offset + size < second_item + 1) {
		    second_item = offset + size - 1;
		}
		memset(set + item - offset, '\1', second_item - item + 1);
	    } else {
		if (offset < item + 1 && item < offset + size) {
		    set[item - offset] = 1;
		}
	    }
	}
	if (!ing) {
	    for (; length < size; length++) {
		ing += set[length];
	    }
	}
    }

    return ing;
}


static int
read_buffer(void *buffer, unsigned size)
{
    int ing;

    ing = read(fd, buffer, size);
    if (ing < 0) {
	longjmp(env, READ_FAILURE);
    } else if (!ing) {
	longjmp(env, UNEXPECTED_EOF);
    }

    return ing;
}


static int
read_hexadecimal(unsigned char *string, unsigned *start_code)
{
    char c;

    c = *string;
    if (47 < c && c < 58) {
	c -= 48;
    } else if (64 < c && c < 71) {
	c -= 55;
    } else if (96 < c && c < 103) {
	c -= 87;
    } else {
	return -1;
    }
    *start_code = c << 4;
    c = string[1];
    if (47 < c && c < 58) {
	c -= 48;
    } else if (64 < c && c < 71) {
	c -= 55;
    } else if (96 < c && c < 103) {
	c -= 87;
    } else {
	return -1;
    }
    *start_code |= c;

    return 0;
}


static int
select_system(unsigned *file_offset, unsigned *first_start_code,
	      unsigned *system_type)
{
    int error;
    unsigned char storage[VIDEO_BUFFER_SIZE];
    struct seek_h262_whatever_type h262_whatever;

    h262_whatever.abort_decoding = error_exit_0;
    h262_whatever.buffer = NULL;
    h262_whatever.correct_underflow = correct_underflow_0;
    h262_whatever.minimum_useful_size = 64;
    h262_whatever.size = 0;

    stream_excess = 0;
    stream_storage = storage;

    if ((error = setjmp(env))) {
	return IDENTIFICATION_FAILURE;
    } else {
	seek_h262_next_start_code(&h262_whatever);
	*first_start_code = h262_whatever.start_code;
	*file_offset = lseek(fd, 0, SEEK_CUR);
	if ((int) *file_offset == -1) {
	    return SEEK_FAILURE;
	}
	*file_offset -= h262_whatever.size + 4;
	if (*first_start_code == pack_start_code) {
	    if (!h262_whatever.size) {
		h262_whatever.correct_underflow(&h262_whatever);
	    }
	    if ((h262_whatever.buffer[0] & 0xf0) == 0x20) {
	    } else if ((h262_whatever.buffer[0] & 0xc0) == 0x40) {
		*system_type = 13818;
	    } else {
		return IDENTIFICATION_FAILURE;
	    }
	} else if (*first_start_code == sequence_header_code) {
	    *system_type = 0;
	} else {
	    *first_start_code = 0;
	}
	if (lseek(fd, *file_offset, SEEK_SET) == -1) {
	    return SEEK_FAILURE;
	}
    }
    if (!error) {
	if (!*first_start_code) {
	    return IDENTIFICATION_FAILURE;
	}
    }

    return error;
}


static unsigned
file_offset(void)
{
    unsigned ing;

    ing = lseek(fd, 0, SEEK_CUR);

    return ing;
}


static void
alloc_storage(void *seek_video, unsigned char **buffers,
	      unsigned luminance_size, unsigned chrominance_size)
{
    unsigned size = 0;

    if (color_space == SEEK_H262_YCbCr) {
	size = (luminance_size + (chrominance_size << 1)) * 3;
    } else if (color_space == SEEK_H262_GRAYSCALE) {
	size = luminance_size * 3;
    }
    if (size) {
	if (!(video_buffer = malloc(size))) {
	    longjmp(env, ALLOCATION_FAILURE);
	}
	if (color_space == SEEK_H262_YCbCr) {
	    unsigned char *buffer;

	    buffer = video_buffer;
	    buffers[0] = buffer;
	    buffer += chrominance_size;
	    buffers[1] = buffer;
	    buffer += chrominance_size;
	    buffers[2] = buffer;
	    buffer += luminance_size;
	    buffers[3] = buffer;
	    buffer += chrominance_size;
	    buffers[4] = buffer;
	    buffer += chrominance_size;
	    buffers[5] = buffer;
	    buffer += luminance_size;
	    buffers[6] = buffer;
	    buffer += chrominance_size;
	    buffers[7] = buffer;
	    buffer += chrominance_size;
	    buffers[8] = buffer;
	} else if (color_space == SEEK_H262_GRAYSCALE) {
	    unsigned char *buffer;

	    buffer = video_buffer;
	    buffers[2] = buffer;
	    buffer += luminance_size;
	    buffers[5] = buffer;
	    buffer += luminance_size;
	    buffers[8] = buffer;
	}
    }
}


static void
correct_underflow_0(void *seek_whatever)
{
    struct seek_h262_whatever_type *whatever;

    whatever = (struct seek_h262_whatever_type *) seek_whatever;
    if (stream_excess) {
	if (whatever->size < stream_excess + 1) {
	    longjmp(env, UNEXPECTED_EOF);
	} else {
	    unsigned excess;

	    if (whatever->buffer != stream_storage) {
		memmove(stream_storage, whatever->buffer, whatever->size);
		whatever->buffer = stream_storage;
	    }

	    excess = whatever->minimum_useful_size - whatever->size;
	    memset(whatever->buffer + whatever->size, 0, excess);
	    whatever->size = whatever->minimum_useful_size;
	    stream_excess += excess;
	}
    } else {
	if (!whatever->size) {
	    int read_size;

	    read_size = read(fd, stream_storage, VIDEO_BUFFER_SIZE);
	    if (read_size < 0) {
		longjmp(env, READ_FAILURE);
	    }
	    whatever->size = read_size;
	} else {
	    int read_size;

	    memmove(stream_storage, whatever->buffer, whatever->size);
	    read_size = read(fd, stream_storage + whatever->size,
			     VIDEO_BUFFER_SIZE - whatever->size);
	    if (read_size < 0) {
		longjmp(env, READ_FAILURE);
	    }
	    whatever->size += read_size;
	}
	whatever->buffer = stream_storage;
	if (whatever->size < whatever->minimum_useful_size) {
	    stream_excess = whatever->minimum_useful_size - whatever->size;
	    memset(whatever->buffer + whatever->size, 0, stream_excess);
	    whatever->size = whatever->minimum_useful_size;
	}
    }
}


static void
correct_underflow_2(void *seek_whatever)
{
    struct seek_h262_whatever_type *whatever;

    whatever = (struct seek_h262_whatever_type *) seek_whatever;
    if (!whatever->size) {
	system_excess = 64 - read_buffer(system_storage, 64);
	whatever->size = 64;
	system_offset = 64;
	whatever->buffer = system_storage;
    } else {
	unsigned this_read;

	if (STORAGE_SIZE < system_offset + 64) {
	    memcpy(system_storage, whatever->buffer, whatever->size);
	    system_offset = whatever->size;
	    whatever->buffer = system_storage;
	}
	this_read = read_buffer(system_storage + system_offset, 64);
	if (this_read + whatever->size < 64) {
	    system_excess = 64 - (whatever->size + this_read);
	    this_read += system_excess;
	} else {
	    system_excess = 0;
	}
	system_offset += this_read;
	whatever->size += this_read;
    }
}


static void
correct_underflow_3(void *seek_whatever)
{
    struct seek_h262_whatever_type *whatever;

    whatever = (struct seek_h262_whatever_type *) seek_whatever;
    if (stream_excess) {
	if (whatever->size < stream_excess + 1) {
	    longjmp(env, UNEXPECTED_EOF);
	} else {
	    unsigned excess;

	    if (whatever->buffer != stream_storage) {
		memmove(stream_storage, whatever->buffer, whatever->size);
		whatever->buffer = stream_storage;
	    }

	    excess = whatever->minimum_useful_size - whatever->size;
	    memset(whatever->buffer + whatever->size, 0, excess);
	    whatever->size = whatever->minimum_useful_size;
	    stream_excess += excess;
	}
    } else {
	if (swap_buffers) {
	    unsigned buffer_offset, buffer_read;

	    buffer_offset = whatever->size;
	    if (buffer_offset) {
		memcpy(stream_storage, whatever->buffer, buffer_offset);
	    }
	    buffer_read = PES_packet_data_size;
	    if (VIDEO_BUFFER_SIZE - buffer_offset < buffer_read) {
		buffer_read = VIDEO_BUFFER_SIZE - buffer_offset;
		PES_packet_data_size -= buffer_read;
		swap_buffers = 1;
	    } else {
		swap_buffers = 0;
	    }
	    read_packet(stream_storage + buffer_offset, buffer_read);
	    whatever->buffer = stream_storage;
	    whatever->size = buffer_read + buffer_offset;
	}
	if (whatever->size < whatever->minimum_useful_size) {
	    struct seek_h262_whatever_type *system;
	    unsigned buffer_start, buffer_offset;

	    buffer_offset = whatever->size;
	    if (buffer_offset) {
		buffer_start = whatever->buffer - stream_storage;
		buffer_offset += buffer_start;
	    } else {
		buffer_start = 0;
		whatever->buffer = stream_storage;
	    }

	    system = (struct seek_h262_whatever_type *) h262_system;
	    while (1) {
		if (!have_start_code) {
		    seek_h262_next_start_code(h262_system);
		}
		if (system->start_code == pack_start_code) {
		    pack_header(h262_system);
		    have_start_code = 1;
		} else if (system_header_start_code < system->start_code) {
		    packet(h262_system);
		    have_start_code = 0;
		    if (PES_packet_data_size
			&& stream_id(system->start_code) == stream_id) {
			unsigned buffer_read;

			if (system->size) {
			    if (system->size < system_excess) {
				longjmp(env, SYSTEM_FAILURE);
			    } else {
				system->size -= system_excess;
			    }
			    if (lseek(fd, -(int) system->size, SEEK_CUR)
				== -1) {
				longjmp(env, SEEK_FAILURE);
			    }
			    system->size = 0;
			}
			buffer_read = PES_packet_data_size;
			if (lseek(fd, -(int) buffer_read, SEEK_CUR) == -1) {
			    longjmp(env, SEEK_FAILURE);
			}
			if (VIDEO_BUFFER_SIZE < buffer_offset + buffer_read) {
			    if (VIDEO_BUFFER_SIZE - buffer_start
				< whatever->minimum_useful_size) {
				buffer_start = 0;
				buffer_offset = whatever->size;
				memmove(stream_storage, whatever->buffer,
					buffer_offset);
				whatever->buffer = stream_storage;
			    }
			    buffer_read = VIDEO_BUFFER_SIZE - buffer_offset;
			    if (PES_packet_data_size < buffer_read + 1) {
				buffer_read = PES_packet_data_size;
			    } else {
				PES_packet_data_size -= buffer_read;
				swap_buffers = 1;
			    }
			}
			read_packet
			    (stream_storage + buffer_offset, buffer_read);
			buffer_offset += buffer_read;
			whatever->size += buffer_read;
			if (whatever->minimum_useful_size
			    < whatever->size + 1) {
			    break;
			}
		    }
		} else if (system->start_code == MPEG_program_end_code) {
		    if (VIDEO_BUFFER_SIZE - buffer_start
			< whatever->minimum_useful_size) {
			buffer_start = 0;
			buffer_offset = whatever->size;
			memmove
			    (stream_storage, whatever->buffer, buffer_offset);
			whatever->buffer = stream_storage;
		    }
		    stream_excess =
			whatever->minimum_useful_size - whatever->size;
		    memset(whatever->buffer + whatever->size, 0,
			   stream_excess);
		    whatever->size = whatever->minimum_useful_size;
		    have_start_code = 1;
		    break;
		} else {
		    longjmp(env, PACKET_START_CODE_MISMATCH);
		}
	    }
	}
    }
}


static void
do_release_0(void *seek_video, struct seek_h262_release_type *release)
{
}


static void
do_release_1(void *seek_video, struct seek_h262_release_type *release)
{
    unsigned data_size, height, i = 0, nfiles, pad, width;

    file_index++;
    if (((release->flags & 7) == SEEK_H262_bidirectionally_predictive_coded
	 && skip_bidirectionally_predictive_coded)
	|| ((release->flags & 7) == SEEK_H262_predictive_coded
	    && skip_predictive_coded)) {
	return;
    }
    data_size = horizontal_size * vertical_size;
    height = vertical_size;
    nfiles = color_space == SEEK_H262_GRAYSCALE ? 1 : 3;
    width = horizontal_size;
    pad = (16 - (width & 15)) & 15;
    for (; i < nfiles; i++) {
	if (is_system) {
	    sprintf(name + prefix_length, ".%02x", stream_id);
	}
	sprintf(name + prefix_length + is_system,
		!i ? ".%08x.Y.%s.pgm"
		: i == 1 ? ".%08x.Cb.%s.pgm" : ".%08x.Cr.%s.pgm",
		file_index, picture_coding_type_names[release->flags & 7]);
	do {
	    FILE *file;

	    if (!(file = fopen(name, "w"))) {
		if (new_line) {
		    new_line = 0;
		    putchar('\n');
		    fflush(stdout);
		}
		fprintf(stderr, " fail to open `%s': ", name);
		fputs(strerror(errno), stderr);
		break;
	    }
	    if (i == 1) {
		if (chroma_format == SEEK_H262_H1V1) {
		    pad = (16 - (width & 15)) & 15;
		} else if (chroma_format == SEEK_H262_H2V1) {
		    data_size >>= 1;
		    width >>= 1;
		    pad = (8 - (width & 7)) & 7;
		} else {
		    data_size >>= 2;
		    height >>= 1;
		    width >>= 1;
		    pad = (8 - (width & 7)) & 7;
		}
	    }
	    fprintf(file, "P5\n%u %u\n255\n", width, height);
	    if (pad) {
		unsigned ii, line_size;
		unsigned char *data;

		data = !i ? release->Y : i == 1 ? release->Cb : release->Cr;
		line_size = width + pad;
		for (ii = height; ii; ii--) {
		    fwrite(data, width, 1, file);
		    data += line_size;
		}
	    } else {
		fwrite(!i ? release->Y : i == 1 ? release->Cb : release->Cr,
		       data_size, 1, file);
	    }
	    fclose(file);
	} while (0);
    }
}


static void
error_exit_0(void *bull_0)
{
    longjmp(env, SYSTEM_FAILURE);
}


static void
error_exit_2(void *bull_0)
{
    int stat_error = 0;

    switch (ERROR(h262_video)) {
    case SEEK_H262_VIDEO_CHANGED_SEQUENCE_HEADER:
	break;
    case SEEK_H262_VIDEO_INVALID_11172_PICTURE_HEADER_INFORMATION:
	stat_error = 1;
	break;
    case SEEK_H262_VIDEO_INVALID_F_CODES:
	break;
    case SEEK_H262_VIDEO_MISSING_SEQUENCE_DISPLAY_EXTENSION:
    case SEEK_H262_VIDEO_MISSING_SEQUENCE_SCALABLE_EXTENSION:
    case SEEK_H262_VIDEO_MISS_F_CODE_0__0_:
    case SEEK_H262_VIDEO_MISS_F_CODE_0__1_:
	stat_error = 1;
	break;
    default:
	longjmp(env, VIDEO_FAILURE);
    }

    if (stat_error) {
	if (new_line) {
	    new_line = 0;
	    fputc('\n', stdout);
	    fflush(stdout);
	}
	stat_video
	    (argv_0, h262_video, file_name, file_offset() - SIZE(h262_video));
    }
}


static void
packet_data(void *seek_whatever, unsigned packet_length)
{
    struct seek_h262_whatever_type *whatever;

    whatever = (struct seek_h262_whatever_type *) seek_whatever;

    if (packet_length < whatever->size) {
	whatever->buffer += packet_length;
	whatever->size -= packet_length;
    } else {
	system_offset = 0;
	packet_length -= whatever->size;
	if (packet_length) {
	    if ((off_t) lseek(fd, packet_length, SEEK_CUR) == -1) {
		longjmp(env, SEEK_FAILURE);
	    }
	}
	whatever->size = 0;
    }
}


static void
read_packet(void *buffer, unsigned size)
{
    unsigned transfer;

    while (size) {
	transfer = read(fd, buffer, size);
	if ((int) transfer < 0) {
	    longjmp(env, READ_FAILURE);
	} else if (!transfer) {
	    longjmp(env, UNEXPECTED_EOF);
	} else {
	    buffer += transfer;
	    size -= transfer;
	}
    }
}


static void
seek_stream(void *seek_whatever, unsigned offset)
{
    struct seek_h262_whatever_type *whatever;

    whatever = (struct seek_h262_whatever_type *) seek_whatever;

    system_offset = 0;
    offset -= whatever->size;
    if ((off_t) lseek(fd, offset, SEEK_CUR) == -1) {
	longjmp(env, SEEK_FAILURE);
    }
    whatever->size = 0;
}


static void
stat_failure(int error)
{
    switch (error) {
    case ALLOCATION_FAILURE:
	fprintf(stderr, "%s: fail to allocate frame store memory for `%s'\n",
		argv_0, file_name);
	break;
    case IDENTIFICATION_FAILURE:
	fprintf(stderr, "%s: I am not so sure that `%s' is indeed a MPEG2_pr"
		"ogram_stream, a iso11172_stream or a video_sequence.\n",
		argv_0, file_name);
	break;
    case MALLOC_FAILURE:
	fprintf(stderr, "%s: cannot allocate memory for `%s' operations\n",
		argv_0, file_name);
	break;
    case PACKET_START_CODE_MISMATCH:
	fprintf(stderr, "%s: I was expecting a packet_start_code at %08x in "
		"`%s', instead I found %08x (shit?).\n", argv_0,
		file_offset() - SIZE(h262_system) - 4, file_name,
		START_CODE(h262_system));
	break;
    case READ_FAILURE:
	fprintf(stderr, "%s: cannot read `%s'\n", argv_0, file_name);
	break;
    case SEQUENCE_SCALABLE_EXTENSION_SEEN:
	fprintf(stderr, "%s: scalable bitstream (as `%s') are not supported\n",
		argv_0, file_name);
	break;
    case SEEK_FAILURE:
	fprintf(stderr, "%s: cannot seek `%s'\n", argv_0, file_name);
	break;
    case SYSTEM_FAILURE:
	stat_system(argv_0, h262_system, file_name,
		    file_offset() - SIZE(h262_system));
	break;
    case UNEXPECTED_EOF:
	fprintf(stderr, "%s: unexpected EOF in `%s' at %08x\n", argv_0,
		file_name, file_offset());
	break;
    case UNEXPECTED_START_CODE:
	fprintf(stderr, "%s: unexpected start code in `%s': %08x\n", argv_0,
		file_name, START_CODE(h262_system));
	break;
    case VIDEO_FAILURE:
	stat_video
	    (argv_0, h262_video, file_name, file_offset() - SIZE(h262_video));
    }
}


static void
stat_progress(void)
{
    double time;
    struct timeval current_time;

    gettimeofday(&current_time, NULL);
    time = current_time.tv_sec - start_time.tv_sec;
    time += (double) (current_time.tv_usec - start_time.tv_usec) / 1000000;
    printf("\rframe, second, rate: %7u, %5u, %9.2f", temporal_reference,
	   (unsigned) time, (double) temporal_reference / time);
    fflush(stdout);
    new_line = 1;
}


static void
stat_system(char *argv_0, void *h262_system, char *file_name,
	    unsigned file_offset)
{
    char *error_string, *start_code_name;

    seek_h262_start_code_name
	(start_code(START_CODE(h262_system)), &start_code_name);
    seek_h262_system_error_string(ERROR(h262_system), &error_string);
    fprintf(stderr, "%s: system parse error in `%s' at %08x (%d) (last decoded"
	    " start code: 000001%02x (%s), %s)\n", argv_0, file_name,
	    file_offset, file_offset, start_code(START_CODE(h262_system)),
	    start_code_name, error_string);
}


static void
stat_video(char *argv_0, void *h262_video, char *file_name,
	   unsigned file_offset)
{
    char *error_string, *start_code_name;

    seek_h262_video_error_string(ERROR(h262_video), &error_string);
    seek_h262_start_code_name
	(start_code(START_CODE(h262_video)), &start_code_name);
    fprintf(stderr, "%s: video parse error in `%s' at %08x (%d) (last decoded "
	    "start code: 000001%02x (%s), %s)\n", argv_0, file_name,
	    file_offset, file_offset, start_code(START_CODE(h262_video)),
	    start_code_name, error_string);
}


static void
x00000100___hack_1_0(struct seek_h262_x00000100___type *picture_header)
{
    if (color_space != SEEK_H262_none || 0) {
	int set_color = 0, set_none = 0;

	if (picture_header->picture_coding_type
	    == SEEK_H262_bidirectionally_predictive_coded) {
	    if (skip_bidirectionally_predictive_coded
		&& temporar_color_space != SEEK_H262_none) {
		set_none = 1;
	    }
	} else if (picture_header->picture_coding_type
		   == SEEK_H262_predictive_coded) {
	    if (skip_predictive_coded) {
		if (temporar_color_space != SEEK_H262_none) {
		    set_none = 1;
		}
	    } else {
		if (temporar_color_space == SEEK_H262_none) {
		    set_color = 1;
		}
	    }
	} else {
	    if (temporar_color_space == SEEK_H262_none) {
		set_color = 1;
	    }
	}
	if (set_none) {
	    seek_h262_set_colorspace(h262_video, SEEK_H262_none);
	    temporar_color_space = SEEK_H262_none;
	} else if (set_color) {
	    seek_h262_set_colorspace(h262_video, color_space);
	    temporar_color_space = color_space;
	}
    }

    temporal_reference++;
}


static void
x00000100___hack_1_2(struct seek_h262_x00000100___type *picture_header)
{
    if (!(temporal_reference & 31)) {
	stat_progress();
    }
    x00000100___hack_1_0(picture_header);
}


static void
x00000100___hack_1_3(struct seek_h262_x00000100___type *picture_header)
{
    temporal_reference++;
}


static void
x000001b5_1_hack_1_1(struct seek_h262_x000001b5_1_type *sequence_extension)
{
    chroma_format = sequence_extension->chroma_format;
    horizontal_size = sequence_extension->horizontal_size;
    vertical_size = sequence_extension->vertical_size;
    x000001b5_1_hack_1_2(sequence_extension);
}


static void
x000001b5_1_hack_1_2(struct seek_h262_x000001b5_1_type *sequence_extension)
{
#if __INTERFACE_SELECTOR__
    synthetic_sequence_extension = sequence_extension->synthetic;
#endif				/* __INTERFACE_SELECTOR__ */
    if (sequence_extension->synthetic) {
    } else if (color_space == SEEK_H262_YCbCr) {
	seek_h262_set_colorspace(h262_video, SEEK_H262_YCbCr);
    }
}


static void
x000001b5_5_hack_0(void *seek_whatever)
{
    longjmp(env, SEQUENCE_SCALABLE_EXTENSION_SEEN);
}


static void
x000001ff___hack_2_1(struct seek_h262_x000001ff___type *PES_packet)
{
    PES_packet_data_size = PES_packet->x.x1.data_size;
}


static void
usage(void)
{
    puts("Usage: h262 [OPTIONS] FILE\n\
Decode system or video streams in FILE(s).\n\
\n\
Options:\n\
  -B, --skip-bidirectionally-predictive-coded\n\
				skip bidirectionally predictive coded pictur\
es\n\
  -I, --idct dummy|mpeg3video_idct_conversion\n\
				select inverse dct routine\n\
  -P, --skip-predictive-coded	skip predictive coded pictures (implies -B)"
#if __Berkeley_FUCK_UP__ == 1
	 "\n\
  -b, --Berkeley-fuck-up	fuck predictions the Berkeley way"
#endif				/* __Berkeley_FUCK_UP__ == 1 */
	 "\n\
  -c, --color-space GRAYSCALE|YCbCr|none\n\
				set output color space (YCbCr by default)\n\
  -p, --ppm=PREFIX		?\n\
  -s, --streams STREAMS		specify the video streams that are to be\n\
				decoded (e0 most of the times)\n\
  -t, --timing			output decoding time\n\
  -v, --verbose			output decoding progress\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    char *color_space_specs = NULL, *idct_conversion_specs = NULL;
    unsigned nfiles;
    unsigned char *target_video_streams = NULL;

    unsetenv("POSIXLY_CORRECT");

    while (1) {
	char c;
	static struct option long_options[] = {
/* *INDENT-OFF* */
#if __Berkeley_FUCK_UP__ == 1
	    {	"Berkeley-fuck-up",
				0x00,	NULL,	0x62	},
#endif				/* __Berkeley_FUCK_UP__ == 1 */
	    {	"color-space",	0x01,	NULL,	0x63	},
	    {	"idct",		0x01,	NULL,	0x49	},
	    {	"ppm",		0x02,	NULL,	0x70	},
	    {	"skip-bidirectionally-predictive-coded",
				0x00,	NULL,	0x42	},
	    {	"skip-predictive-coded",
				0x00,	NULL,	0x50	},
	    {	"streams",	0x01,	NULL,	0x73	},
	    {	"timing",	0x00,	NULL,	0x74	},
	    {	"verbose",	0x00,	NULL,	0x76	},
	    {	"version",	0x00,	NULL,	0x56	},
	    {	"help",		0x00,	NULL,	0x48	},
	    {	NULL,		0x00,	NULL,	0x00	}
/* *INDENT-ON* */
	};

/* *INDENT-OFF* */
	c = getopt_long(argc, argv, "BI:P"
#if __Berkeley_FUCK_UP__ == 1
					"b"
#endif				/* __Berkeley_FUCK_UP__ == 1 */
					 "c:p::s:tv", long_options, NULL);
/* *INDENT-ON* */

	if (!~c) {
	    break;
	}

	switch (c) {
	case 'B':
	    skip_bidirectionally_predictive_coded = 1;
	    break;
	case 'I':
	    idct_conversion_specs = optarg;
	    break;
	case 'P':
	    skip_bidirectionally_predictive_coded = 1;
	    skip_predictive_coded = 1;
	    break;
#if __Berkeley_FUCK_UP__ == 1
	case 'b':
	    Berkeley_fuck_up = 1;
	    break;
#endif				/* __Berkeley_FUCK_UP__ == 1 */
	case 'c':
	    color_space_specs = optarg;
	    break;
	case 'p':
	    do_ppm = 1;
	    if (optarg) {
		prefix = optarg;
	    }
	    break;
	case 's':
	    target_video_streams = (unsigned char *) optarg;
	    break;
	case 't':
	    do_time = 1;
	    break;
	case 'v':
	    do_progress = do_time = 1;
	    break;
	case 'H':
	    usage();

	    return 0;
	case 'V':
	    printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);

	    return 0;
	case '?':
	    return 1;
	}
    }

    if (argc < 2) {
	fprintf(stderr, "%s: too few arguments\nTry `%s --help' for more inf"
		"ormation\n", argv[0], argv[0]);

	return 1;
    }

    if ((int) (nvideo_streams = parse_set
	       (video_streams, 0xe0, 0x10, target_video_streams)) == -1) {
	fprintf(stderr, "%s: error parsing video_streams specification\n",
		argv[0]);

	return 1;
    }

    if (color_space_specs) {
	if (!find_name(color_spaces, color_space_specs, &color_space)) {
	    fprintf(stderr, "%s: cannot parse color space specification: `%s"
		    "'\n", argv[0], color_space_specs);

	    return 1;
	}
    }

    if (do_ppm && color_space == SEEK_H262_none) {
	fprintf(stderr, "%s: if ppm output is what you want you must specify"
		" the desired color space (other than none)\n", argv[0]);

	return 1;
    }

    if (idct_conversion_specs) {
	int index;

	if (!find_name(idct_routines, idct_conversion_specs, &index)) {
	    fprintf(stderr, "%s: cannot parse idct specification: `%s'\n",
		    argv[0], idct_conversion_specs);

	    return 1;
	}

	idct_add_method = idcts[index].idct_add_block;
	idct_set_method = idcts[index].idct_set_block;
	init_video_idct = idcts[index].init_idct;
    } else {
	idct_add_method = mpeg3video_idct_conversion_0;
	idct_set_method = mpeg3video_idct_conversion_1;
	init_video_idct = mpeg3video_idct_conversion_3;
    }

    temporar_color_space = color_space;

    argv_0 = argv[0];

    nfiles = argc - optind;
    for (; optind < argc; optind++) {
	int error;
	unsigned file_offset = 0, first_start_code = 0, system_type = 11172;

	file_name = argv[optind];
	if ((do_progress || do_time) && nfiles != 1) {
	    printf("%s:\n", file_name);
	}

	if ((fd = open(file_name, O_RDONLY)) == -1) {
	    fprintf(stderr, "%s: fail to open `%s'\n", argv[0], file_name);
	    continue;
	}

	error = select_system(&file_offset, &first_start_code, &system_type);
	if (!error) {
	    if (first_start_code == pack_start_code && !nvideo_streams) {
		error = decode_system(system_type);
	    } else {
		error = decode_stream
		    (file_offset, first_start_code, system_type);
	    }
	}

	if (error) {
	    stat_failure(error);
	}

	close(fd);
    }

    return 0;
}
